from compas_slicer.print_organization import BasePrintOrganizer
from compas_slicer.pre_processing.preprocessing_utils import topological_sorting as topo_sort
from compas_slicer.print_organization.curved_print_organization import BaseBoundary
import compas_slicer
from compas.geometry import closest_point_on_polyline, distance_point_point, Polyline, Vector, normalize_vector, Point
import logging
from compas_slicer.geometry import Path, PrintPoint
import compas_slicer.utilities as utils
from compas_slicer.parameters import get_param

logger = logging.getLogger('logger')

__all__ = ['InterpolationPrintOrganizer']


class InterpolationPrintOrganizer(BasePrintOrganizer):
    """
    Organizing the printing process for the realization of non-planar contours.

    Attributes
    ----------
    slicer: :class:`compas_slicer.slicers.PlanarSlicer`
        An instance of the compas_slicer.slicers.PlanarSlicer.
    parameters: dict
    DATA_PATH: str
    """

    def __init__(self, slicer, parameters, DATA_PATH):
        assert isinstance(slicer, compas_slicer.slicers.InterpolationSlicer), 'Please provide an InterpolationSlicer'
        BasePrintOrganizer.__init__(self, slicer)
        self.DATA_PATH = DATA_PATH
        self.OUTPUT_PATH = utils.get_output_directory(DATA_PATH)
        self.parameters = parameters

        self.vertical_layers = slicer.vertical_layers
        self.horizontal_layers = slicer.horizontal_layers
        assert len(self.vertical_layers) + len(self.horizontal_layers) == len(slicer.layers)

        if len(self.horizontal_layers) > 0:
            assert len(self.horizontal_layers) == 1, "Only one brim horizontal layer is currently supported."
            assert self.horizontal_layers[0].is_brim, "Only one brim horizontal layer is currently supported."
            logger.info('Slicer has one horizontal brim layer.')

        # topological sorting of vertical layers depending on their connectivity
        self.topo_sort_graph = None
        if len(self.vertical_layers) > 1:
            self.topological_sorting()
        self.selected_order = None

        # creation of one base boundary per vertical_layer
        self.base_boundaries = self.create_base_boundaries()

    def __repr__(self):
        return "<InterpolationPrintOrganizer with %i vertical_layers>" % len(self.vertical_layers)

    def topological_sorting(self):
        """ When the print consists of various paths, this function initializes a class that creates
        a directed graph with all these parts, with the connectivity of each part reflecting which
        other parts it lies on, and which other parts lie on it."""
        max_layer_height = get_param(self.parameters, key='max_layer_height', defaults_type='layers')
        self.topo_sort_graph = topo_sort.SegmentsDirectedGraph(self.slicer.mesh, self.vertical_layers,
                                                               max_layer_height, DATA_PATH=self.DATA_PATH)

    def create_base_boundaries(self):
        """ Creates one BaseBoundary per vertical_layer."""
        bs = []
        root_vs = utils.get_mesh_vertex_coords_with_attribute(self.slicer.mesh, 'boundary', 1)
        root_boundary = BaseBoundary(self.slicer.mesh, [Point(*v) for v in root_vs])

        if len(self.vertical_layers) > 1:
            for i, vertical_layer in enumerate(self.vertical_layers):
                parents_of_current_node = self.topo_sort_graph.get_parents_of_node(i)
                if len(parents_of_current_node) == 0:
                    boundary = root_boundary
                else:
                    boundary_pts = []
                    for parent_index in parents_of_current_node:
                        parent = self.vertical_layers[parent_index]
                        boundary_pts.extend(parent.paths[-1].points)
                    boundary = BaseBoundary(self.slicer.mesh, boundary_pts)
                bs.append(boundary)
        else:
            bs.append(root_boundary)

        # save intermediary outputs
        b_data = {i: b.to_data() for i, b in enumerate(bs)}
        utils.save_to_json(b_data, self.OUTPUT_PATH, 'boundaries.json')

        return bs

    def create_printpoints(self):
        """
        Create the print points of the fabrication process
        Based on the directed graph, select one topological order.
        From each path collection in that order copy PrintPoints dictionary in the correct order.
        """
        current_layer_index = 0

        # (1) --- First add the printpoints of the horizontal brim layer (first layer of print)
        self.printpoints_dict['layer_0'] = {}
        if len(self.horizontal_layers) > 0:  # first add horizontal brim layers
            paths = self.horizontal_layers[0].paths
            for j, path in enumerate(paths):
                self.printpoints_dict['layer_0']['path_%d' % j] = \
                    [PrintPoint(pt=point, layer_height=get_param(self.parameters, 'avg_layer_height', 'layers'),
                                mesh_normal=utils.get_normal_of_path_on_xy_plane(k, point, path, self.slicer.mesh))
                     for k, point in enumerate(path.points)]
            current_layer_index += 1

        # (2) --- Select order of vertical layers
        if len(self.vertical_layers) > 1:  # the you need to select one topological order
            all_orders = self.topo_sort_graph.get_all_topological_orders()
            self.selected_order = all_orders[0]  # TODO: add more elaborate selection strategy
        else:
            self.selected_order = [0]  # there is only one segment, only this option

        # (3) --- Then create the printpoints of all the vertical layers in the selected order
        for index, i in enumerate(self.selected_order):
            layer = self.vertical_layers[i]
            self.printpoints_dict['layer_%d' % current_layer_index] = {}
            self.printpoints_dict['layer_%d' % current_layer_index] = self.get_layer_ppts(layer, self.base_boundaries[i])
            current_layer_index += 1

    def get_layer_ppts(self, layer, base_boundary):
        """ Creates the PrintPoints of a single layer."""
        max_layer_height = get_param(self.parameters, key='max_layer_height', defaults_type='layers')
        min_layer_height = get_param(self.parameters, key='min_layer_height', defaults_type='layers')
        avg_layer_height = get_param(self.parameters, 'avg_layer_height', 'layers')

        all_pts = [pt for path in layer.paths for pt in path.points]
        closest_fks, projected_pts = utils.pull_pts_to_mesh_faces(self.slicer.mesh, all_pts)
        normals = [Vector(*self.slicer.mesh.face_normal(fkey)) for fkey in closest_fks]

        count = 0
        crv_to_check = Path(base_boundary.points, True)  # creation of fake path for the lower boundary

        layer_ppts = {}
        for i, path in enumerate(layer.paths):
            layer_ppts['path_%d' % i] = []

            for p in path.points:
                cp = closest_point_on_polyline(p, Polyline(crv_to_check.points))
                d = distance_point_point(cp, p)

                ppt = PrintPoint(pt=p, layer_height=avg_layer_height, mesh_normal=normals[count])

                ppt.closest_support_pt = Point(*cp)
                ppt.distance_to_support = d
                ppt.layer_height = max(min(d, max_layer_height), min_layer_height)
                ppt.up_vector = Vector(*normalize_vector(Vector.from_start_end(cp, p)))
                ppt.frame = ppt.get_frame()

                layer_ppts['path_%d' % i].append(ppt)
                count += 1

            crv_to_check = path

        return layer_ppts


if __name__ == "__main__":
    pass



import os
import logging
import numpy as np
import copy
import compas
import compas_slicer.utilities as utils
from compas_slicer.pre_processing.preprocessing_utils import restore_mesh_attributes, save_vertex_attributes
from compas.datastructures import Mesh
from compas_slicer.pre_processing.preprocessing_utils import assign_interpolation_distance_to_mesh_vertex
from compas_slicer.slicers.slice_utilities import ScalarFieldContours
from compas_slicer.pre_processing.preprocessing_utils import assign_interpolation_distance_to_mesh_vertices
from compas_slicer.pre_processing.gradient_evaluation import GradientEvaluation
from compas.geometry import Line, distance_point_point_sqrd, project_point_line

packages = utils.TerminalCommand('conda list').get_split_output_strings()
if 'igl' in packages:
    import igl

logger = logging.getLogger('logger')

__all__ = ['MeshSplitter']

# --- Parameters
T_SEARCH_RESOLUTION = 60000
HIT_THRESHOLD = 0.02


class MeshSplitter:
    """
    Curved slicing pre-processing step.

    Takes one continuous mesh with various saddle points and splits it up
    at every saddle point following the direction of the iso-contour that intersects that saddle point,
    so that the resulting mesh has no remaining saddle points.

    The result is a series of split meshes whose vertex attributes have been updated with boundary attributes at the
    newly created cuts, (i.e. they all have vertex 'boundary' attributes 1,2 on their lower and upper
    boundaries)

    For each newly created mesh, a separate slicer needs to be created. Like that,
    we will always have one slicer per mesh with the correct attributes already assigned.
    However, it can still happen that a slicer that takes a split mesh outputs more than one
    vertical_layers_print_data (vertical layers).

    Attributes
    ----------
    mesh: :class: 'compas.datastructures.Mesh'
    target_LOW: :class: 'compas_slicer.pre_processing.CompoundTarget'
    target_HIGH: :class: 'compas_slicer.pre_processing.CompoundTarget'
    DATA_PATH: str, the path to the data folder
    """

    def __init__(self, mesh, target_LOW, target_HIGH, DATA_PATH):
        self.mesh = mesh  # compas mesh
        self.DATA_PATH = DATA_PATH
        self.OUTPUT_PATH = utils.get_output_directory(DATA_PATH)
        self.target_LOW, self.target_HIGH = target_LOW, target_HIGH

        assign_interpolation_distance_to_mesh_vertices(self.mesh, weight=0.5, target_LOW=self.target_LOW,
                                                       target_HIGH=self.target_HIGH)
        g_evaluation = GradientEvaluation(self.mesh, self.DATA_PATH)
        g_evaluation.find_critical_points()  # First estimation of saddle points with weight = 0.5
        self.saddles = g_evaluation.saddles
        self.cut_indices = []

    # --------------------------- main
    def run(self):
        """
        Runs the mesh splitting process. This consists of the following parts.

        (1) Find the iso-contours that intersect the saddle points.
        Iteratively find the weights (from 0 to 1) that output a distance field whose iso-contour is intersecting
        each saddle point. Here two iterations are carried out (one rough and one exact search).

        For each saddle point and its respective weight and iso-contour:
        (2) Find the zero-crossing points of the iso-contour and merge points that are close to the saddle to ensure
        connection.
        (3) Cleanup iso-contour. Only keep neighborhoods that are intersecting the saddle point. Discard all other.
        (4) Create the cut on the mesh.
        (5) Weld the resulting mesh, and restore all the vertex attributes. Note that the mesh remains in one piece,
        although the cuts have been created.
        (6) Update compound tardets with the new mesh.
        """

        # (1) first rough estimation of split params
        split_params = self.identify_positions_to_split(self.saddles)
        # TODO: merge params that are too close together to avoid creation of very thin neighborhoods.
        logger.info("%d Split params. First rough estimation :  " % len(split_params) + str(split_params))

        # split mesh at params
        logger.info('Splitting mesh at split params')
        current_cut_index = 1

        for i, param_first_estimation in enumerate(split_params):
            print('')
            logger.info('cut_index : %d, param_first_estimation : %.6f' % (current_cut_index, param_first_estimation))

            # --- (1) More exact estimation of intersecting weight. Recompute gradient evaluation.
            # Find exact saddle point and the weight that intersects it.

            assign_interpolation_distance_to_mesh_vertices(self.mesh, weight=param_first_estimation,
                                                           target_LOW=self.target_LOW, target_HIGH=self.target_HIGH)
            g_evaluation = GradientEvaluation(self.mesh, self.DATA_PATH)
            g_evaluation.find_critical_points()
            saddles_ds_tupples = [(vkey, abs(g_evaluation.mesh.vertex_attribute(vkey, 'scalar_field'))) for vkey in
                                  g_evaluation.saddles]
            saddles_ds_tupples = sorted(saddles_ds_tupples, key=lambda saddle_tupple: saddle_tupple[1])
            vkey = saddles_ds_tupples[0][0]
            t = self.identify_positions_to_split([vkey])[0]
            logger.info('vkey_exact : %d , t_exact : %.6f' % (vkey, t))

            # --- (2) find zero-crossing points
            assign_interpolation_distance_to_mesh_vertices(self.mesh, t, self.target_LOW, self.target_HIGH)
            zero_contours = ScalarFieldContours(self.mesh)
            zero_contours.compute()
            keys_of_clusters_to_keep = merge_clusters_saddle_point(zero_contours, saddle_vkeys=[vkey])

            # --- (3) Cleaning up zero-crossing neighborhoods
            cleanup_unrelated_isocontour_neighborhoods(zero_contours, keys_of_clusters_to_keep)

            if zero_contours:  # if there are remaining zero-crossing neighborhoods
                zero_contours = smoothen_cut(zero_contours, self.mesh, saddle_vkeys=[vkey], iterations=15,
                                             strength=0.2)  # smoothen the cut close to the saddle point.

                # save to json intermediary results
                zero_contours.save_point_clusters_as_polylines_to_json(self.OUTPUT_PATH,
                                                                       'point_clusters_polylines_%d.json' % int(i))

                #  --- (4) Create cut
                logger.info("Creating cut on mesh")
                self.cut_indices.append(current_cut_index)
                self.split_intersected_faces(zero_contours, current_cut_index)
                current_cut_index += 1

                #  --- (5) Weld mesh and restore attributes
                logger.info('Cleaning up the mesh. Welding and restoring attributes')
                v_attributes_dict = save_vertex_attributes(self.mesh)
                self.mesh = weld_mesh(self.mesh, self.OUTPUT_PATH)
                restore_mesh_attributes(self.mesh, v_attributes_dict)

                #  --- (6) Update targets
                if i < len(split_params) - 1:  # does not need to happen at the end
                    logger.info('Updating targets, recomputing geodesic distances')
                    self.update_targets()

    def update_targets(self):
        """
        Update targets with the new mesh that was created during the split process.
        Note: This only works if the target vertices have not been touched. If all has gone well, targets can only have
        minima and maxima, so they should remain intact after the split
        """
        self.target_LOW.assign_new_mesh(self.mesh)
        self.target_LOW.find_targets_connected_components()
        self.target_LOW.compute_geodesic_distances()
        if self.target_HIGH:
            self.target_HIGH.assign_new_mesh(self.mesh)
            self.target_HIGH.find_targets_connected_components()
            self.target_HIGH.compute_geodesic_distances()

    def split_intersected_faces(self, zero_contours, cut_index):
        """
        Create cuts on intersected faces

        Parameters
        ----------
        zero_contours: :class: 'compas_slicer.pre_processing.ScalarFieldContours'
        cut_index: int, the vertex attribute value data['cut'] of the current cut
        """
        for key in zero_contours.sorted_point_clusters:  # cluster_pair
            edges = zero_contours.sorted_edge_clusters[key]
            pts = zero_contours.sorted_point_clusters[key]

            # add first vertex
            p = pts[0]
            v0 = self.mesh.add_vertex(x=p[0], y=p[1], z=p[2], attr_dict={'cut': 1})

            for i, edge in enumerate(edges):
                next_edge = edges[(i + 1) % len(edges)]
                p = pts[(i + 1) % len(pts)]

                faces_current_edge = self.mesh.edge_faces(u=edge[0], v=edge[1])
                faces_next_edge = self.mesh.edge_faces(u=next_edge[0], v=next_edge[1])

                fkey_common = list(set(faces_current_edge).intersection(faces_next_edge))[0]
                vkey_common = list(set(edge).intersection(next_edge))[0]
                v_other_a = list(set(edge).difference([vkey_common]))[0]
                v_other_b = list(set(next_edge).difference([vkey_common]))[0]

                v_new = self.mesh.add_vertex(x=p[0], y=p[1], z=p[2], attr_dict={'cut': cut_index})

                # remove and add faces
                if fkey_common in list(self.mesh.faces()):
                    self.mesh.delete_face(fkey_common)
                    self.mesh.add_face([vkey_common, v_new, v0])
                    self.mesh.add_face([v_new, v_other_a, v0])
                    self.mesh.add_face([v_other_b, v_other_a, v_new])
                else:
                    logger.warning('Did not need to modify faces.')
                v0 = v_new

        self.mesh.cull_vertices()  # remove all unused vertices
        try:
            self.mesh.unify_cycles()
        except AssertionError:
            logger.warning('Could NOT unify cycles')
        if not self.mesh.is_valid():
            logger.warning('Attention! Mesh is NOT valid!')

    def identify_positions_to_split(self, saddles):
        """
        Find the weights that create iso-contours that intersect the saddle points.

        Parameters
        ----------
        saddles: list, int, the vertex keys of the saddle points

        Returns
        ----------
        list, float, the weights from 0 to 1. One for each saddle point.
        """
        split_params = []
        for vkey in saddles:
            param = self.find_weight_intersecting_vkey(vkey, threshold=HIT_THRESHOLD, resolution=T_SEARCH_RESOLUTION)
            split_params.append(param)
        return split_params

    def find_weight_intersecting_vkey(self, vkey, threshold, resolution):
        """
        Find the weights that intersect the vertex.

        Parameters
        ----------
        vkey: list, int, the vertex key to intersect
        threshold: float, the d value below which we consider we have a hit. Should be a very small value
        resolution: int, the resolution of search, should be a value more than 10**4

        Returns
        ----------
        float, the weights from 0 to 1.
        """
        weight_list = get_weights_list(n=resolution, start=0.001, end=0.999)
        # TODO: save next d to avoid re-evaluating
        for i, weight in enumerate(weight_list[:-1]):
            current_d = assign_interpolation_distance_to_mesh_vertex(vkey, weight, self.target_LOW, self.target_HIGH)
            next_d = assign_interpolation_distance_to_mesh_vertex(vkey, weight_list[i + 1], self.target_LOW, self.target_HIGH)
            if abs(current_d) < abs(next_d) and current_d < threshold:
                return weight
        raise ValueError('Could NOT find param for saddle vkey %d!' % vkey)


###############################################
# --- helpers
###############################################

def get_weights_list(n, start=0.03, end=1.0):
    """ Returns a numpy array with n numbers from start to end. """
    return list(np.arange(start=start, stop=end, step=(end - start) / n))


###############################################
# --- Separate disconnected components

def separate_disconnected_components(mesh, attr, values, OUTPUT_PATH):
    """
    Given a mesh with cuts that have already been created, it separates the disconnected
    components using the igl function. Then it welds them and restores their attributes.

    Parameters
    ----------
    mesh: :class: 'compas.datastructures.Mesh'
    attr: str, the key of the vertex attributes that signals the cuts. most likely 'cut'
    values: list, int, the cut indices
    OUTPUT_PATH: str, the path to the output folder

    Returns
    ----------
    list, :class: 'compas.datastructures.Mesh' The resulting split meshes.
    """

    v_attributes_dict = save_vertex_attributes(mesh)

    v, f = mesh.to_vertices_and_faces()
    v, f = np.array(v), np.array(f)

    # --- create cut flags for igl
    cut_flags = []
    for fkey in mesh.faces():
        edges = mesh.face_halfedges(fkey)
        current_face_flags = []
        for fu, fv in edges:
            fu_attr, fv_attr = mesh.vertex_attribute(fu, attr), mesh.vertex_attribute(fv, attr)
            if fu_attr == fv_attr and fu_attr in values:
                current_face_flags.append(1)
            else:
                current_face_flags.append(0)
        cut_flags.append(current_face_flags)
    cut_flags = np.array(cut_flags)
    assert cut_flags.shape == f.shape

    # --- cut mesh
    v_cut, f_cut = igl.cut_mesh(v, f, cut_flags)
    connected_components = igl.face_components(f_cut)

    f_dict = {}
    for i in range(max(connected_components) + 1):
        f_dict[i] = []
    for f_index, f in enumerate(f_cut):
        component = connected_components[f_index]
        f_dict[component].append(f)

    cut_meshes = []
    for component in f_dict:
        cut_mesh = Mesh.from_vertices_and_faces(v_cut, f_dict[component])
        cut_mesh.cull_vertices()
        if len(list(cut_mesh.faces())) > 2:
            cut_mesh.to_obj(os.path.join(OUTPUT_PATH, 'temp.obj'))
            cut_mesh = Mesh.from_obj(os.path.join(OUTPUT_PATH, 'temp.obj'))  # get rid of too many empty keys
            cut_meshes.append(cut_mesh)

    for mesh in cut_meshes:
        restore_mesh_attributes(mesh, v_attributes_dict)

    return cut_meshes


###############################################
# --- saddle points merging


def smoothen_cut(zero_contours, mesh, saddle_vkeys, iterations, strength, distance_threshold=20.0*20.0):
    """ Iterative smoothing of the cut around the saddle point. """
    for _ in range(iterations):
        saddles = [mesh.vertex_coordinates(key) for key in saddle_vkeys]
        count = 0

        for cluster_key in zero_contours.sorted_point_clusters:
            pts = zero_contours.sorted_point_clusters[cluster_key]
            edges = zero_contours.sorted_edge_clusters[cluster_key]
            for i, pt in enumerate(pts):
                if 0.01 < min([distance_point_point_sqrd(pt, s) for s in saddles]) < distance_threshold:
                    count += 1
                    edge = edges[i]
                    prev = pts[i - 1]
                    next_p = pts[(i + 1) % len(pts)]
                    avg = [(prev[0] + next_p[0]) * 0.5, (prev[1] + next_p[1]) * 0.5, (prev[2] + next_p[2]) * 0.5]
                    point = np.array(avg) * strength + np.array(pt) * (1 - strength)
                    line = Line(mesh.vertex_coordinates(edge[0]), mesh.vertex_coordinates(edge[1]))
                    projected_pt = project_point_line(point, line)
                    pts[i] = projected_pt
                    zero_contours.sorted_point_clusters[cluster_key][i] = projected_pt

    return zero_contours


def merge_clusters_saddle_point(zero_contours, saddle_vkeys):
    """
    Merge the points that are on edges that have the saddle point as one of their edges to the saddle point.

    Parameters
    ----------
    zero_contours: :class: 'compas_slicer.pre_processing.ScalarFieldContours'
    saddle_vkeys: list, int, the vertex keys of the current saddle points.
    (Currently this can only be a single saddle point)

    Returns
    ----------
    list, int. The index neighborhoods that are related to the saddle points.
    """
    keys_of_clusters_to_keep = []
    for cluster_key in zero_contours.sorted_edge_clusters:
        edges = zero_contours.sorted_edge_clusters[cluster_key]
        for i, e in enumerate(edges):
            for saddle_vkey in saddle_vkeys:
                if saddle_vkey in e:
                    zero_contours.sorted_point_clusters[cluster_key][i] = \
                        zero_contours.mesh.vertex_coordinates(saddle_vkey)  # merge point with saddle point
                    print('Found edge to merge: ' + str(e))
                    if cluster_key not in keys_of_clusters_to_keep:
                        keys_of_clusters_to_keep.append(cluster_key)

    return keys_of_clusters_to_keep


def cleanup_unrelated_isocontour_neighborhoods(zero_contours, keys_of_clusters_to_keep):
    """
    Remove the neighborhoods of the zero-crossing contour when they are not related to the saddle points.

    Parameters
    ----------
    zero_contours: :class: 'compas_slicer.pre_processing.ScalarFieldContours'
    keys_of_clusters_to_keep: list, int. The index neighborhoods that are related to the saddle points.
    """
    if len(keys_of_clusters_to_keep) == 0:
        logger.error("No common vertex found! Skipping this split_param")
        return None
    else:
        logger.info('keys_of_clusters_to_keep : ' + str(keys_of_clusters_to_keep))
        # empty all other clusters that are not in the matching_pair
        sorted_point_clusters_clean = copy.deepcopy(zero_contours.sorted_point_clusters)
        sorted_edge_clusters_clean = copy.deepcopy(zero_contours.sorted_edge_clusters)
        for key in zero_contours.sorted_point_clusters:
            if key not in keys_of_clusters_to_keep:
                del sorted_point_clusters_clean[key]
                del sorted_edge_clusters_clean[key]

        zero_contours.sorted_point_clusters = sorted_point_clusters_clean
        zero_contours.sorted_edge_clusters = sorted_edge_clusters_clean


########################################################
# --- Mesh welding and sanitizing

def weld_mesh(mesh, OUTPUT_PATH, precision='2f'):
    """ Welds mesh and check that the result is valid. """
    for f_key in mesh.faces():
        if len(mesh.face_vertices(f_key)) < 3:
            mesh.delete_face(f_key)

    welded_mesh = compas.datastructures.mesh_weld(mesh, precision=precision)

    welded_mesh.to_obj(os.path.join(OUTPUT_PATH, 'temp.obj'))  # make sure there's no empty f_keys
    welded_mesh = Mesh.from_obj(os.path.join(OUTPUT_PATH, 'temp.obj'))  # TODO: find a better way to do this

    try:
        welded_mesh.unify_cycles()
        logger.info("Unified cycles of welded_mesh")
    except AssertionError:
        logger.error("Attention! Could NOT unify cycles of welded_mesh")

    if not welded_mesh.is_valid():  # and iteration < 3:
        logger.error("Attention! Welded mesh is INVALID")
    if not welded_mesh.is_manifold():
        logger.error("Attention! Welded mesh is NON-MANIFOLD")

    return welded_mesh


if __name__ == "__main__":
    pass



import logging
from copy import deepcopy

logger = logging.getLogger('logger')

__all__ = ['smooth_printpoint_attribute',
           'smooth_printpoints_up_vectors',
           'smooth_printpoints_layer_heights']


def smooth_printpoint_attribute(print_organizer, iterations, strength, get_attr_value, set_attr_value):
    """
    Iterative smoothing of the printpoints attribute.
    The attribute is accessed using the function 'get_attr_value(ppt)', and is set using the function
    'set_attr_value(ppt, v)'.
    All attributes are smoothened continuously (i.e. as if their printpoints belong into one long uninterrupted path)
    For examples of how to use this function look at 'smooth_printpoints_layer_heights' and
    'smooth_printpoints_up_vectors' below.
    The smoothing is happening by taking an average of the previous and next point attributes, and combining them with
    the current value of the print point; On every iteration:
    new_val = (0.5*(neighbor_left_val + neighbor_right_attr)) * strength - current_val * (1-strength)

    Parameters
    ----------
    print_organizer: :class: 'compas_slicer.print_organization.BasePrintOrganizer', or other class inheriting from it.
    iterations: int, smoothing iterations
    strength: float. in the range [0.0 - 1.0]. 0.0 corresponds to no smoothing at all, 1.0 corresponds to overwriting
        the current value with the average of the two neighbors on every interation stop. On each iteration:
        new_val = (0.5*(neighbor_left_val + neighbor_right_attr)) * strength - current_val * (1-strength)
    get_attr_value: function that returns an attribute of a printpoint, get_attr_value(ppt)
    set_attr_value: function that sets an attribute of a printpoint, set_attr_value(ppt, new_value)
    """

    # first smoothen the values
    for ppt in print_organizer.printpoints_iterator():
        assert get_attr_value(ppt), 'The attribute you are trying to smooth has not been assigned a value'

    attrs = [get_attr_value(ppt) for ppt in print_organizer.printpoints_iterator()]
    new_values = deepcopy(attrs)

    for iteration in range(iterations):
        for i, ppt in enumerate(print_organizer.printpoints_iterator()):
            if 0 < i < len(attrs) - 1:  # ignore first and last element
                mid = (attrs[i - 1] + attrs[i + 1]) * 0.5
                new_values[i] = mid * strength + attrs[i] * (1 - strength)
        attrs = new_values

        # in the end assign the new (smoothened) values to the printpoints
        if iteration == iterations-1:
            for i, ppt in enumerate(print_organizer.printpoints_iterator()):
                set_attr_value(ppt, attrs[i])


def smooth_printpoints_layer_heights(print_organizer, iterations, strength):
    """ This function is an example for how the 'smooth_printpoint_attribute' function can be used. """
    def get_ppt_layer_height(printpoint): return printpoint.layer_height  # get value
    def set_ppt_layer_height(printpoint, v): printpoint.layer_height = v  # set value
    smooth_printpoint_attribute(print_organizer, iterations, strength, get_ppt_layer_height, set_ppt_layer_height)


def smooth_printpoints_up_vectors(print_organizer, iterations, strength):
    """ This function is an example for how the 'smooth_printpoint_attribute' function can be used. """
    def get_ppt_up_vec(printpoint): return printpoint.up_vector  # get value
    def set_ppt_up_vec(printpoint, v): printpoint.up_vector = v  # set value
    smooth_printpoint_attribute(print_organizer, iterations, strength, get_ppt_up_vec, set_ppt_up_vec)
    # finally update any values in the printpoints that are affected by the changed attribute
    for ppt in print_organizer.printpoints_iterator():
        ppt.frame = ppt.get_frame()



__all__ = ['print_organization_default_param']


def print_organization_default_param(key):
    """ Returns the default parameters with the specified key. """
    if key in default_parameters:
        return default_parameters[key]
    else:
        raise ValueError('The parameter with key : ' + str(key) +
                         ' does not exist in the defaults of curved_slicing parameters. ')


default_parameters = \
    {
    }



import logging
from compas_slicer.pre_processing.preprocessing_utils import blend_union_list
from compas_slicer.utilities.utils import remap_unbound

logger = logging.getLogger('logger')

__all__ = ['assign_interpolation_distance_to_mesh_vertices',
           'assign_interpolation_distance_to_mesh_vertex']


def assign_interpolation_distance_to_mesh_vertices(mesh, weight, target_LOW, target_HIGH):
    """
    Fills in the 'get_distance' attribute of every vertex of the mesh.

    Parameters
    ----------
    mesh: :class: 'compas.datastructures.Mesh'
    weight: float,
        The weighting of the distances from the lower and the upper target, from 0 to 1.
    target_LOW: :class: 'compas_slicer.pre_processing.CompoundTarget'
        The lower compound target.
    target_HIGH:  :class: 'compas_slicer.pre_processing.CompoundTarget'
        The upper compound target.
    """
    for i, vkey in enumerate(mesh.vertices()):
        d = assign_interpolation_distance_to_mesh_vertex(vkey, weight, target_LOW, target_HIGH)
        mesh.vertex[vkey]['scalar_field'] = d


def assign_interpolation_distance_to_mesh_vertex(vkey, weight, target_LOW, target_HIGH):
    """
    Fills in the 'get_distance' attribute for a single vertex with vkey.

    Parameters
    ----------
    vkey: int
        The vertex key.
    weight: float,
        The weighting of the distances from the lower and the upper target, from 0 to 1.
    target_LOW: :class: 'compas_slicer.pre_processing.CompoundTarget'
        The lower compound target.
    target_HIGH:  :class: 'compas_slicer.pre_processing.CompoundTarget'
        The upper compound target.
    """
    if target_LOW and target_HIGH:  # then interpolate targets
        d = get_weighted_distance(vkey, weight, target_LOW, target_HIGH)
    elif target_LOW:  # then offset target
        offset = weight * target_LOW.get_max_dist()
        d = target_LOW.get_distance(vkey) - offset
    else:
        raise ValueError('You need to provide at least one target')
    return d


def get_weighted_distance(vkey, weight, target_LOW, target_HIGH):
    """
    Computes the weighted get_distance for a single vertex with vkey.

    Parameters
    ----------
    vkey: int
        The vertex key.
    weight: float,
        The weighting of the distances from the lower and the upper target, from 0 to 1.
    target_LOW: :class: 'compas_slicer.pre_processing.CompoundTarget'
        The lower compound target.
    target_HIGH:  :class: 'compas_slicer.pre_processing.CompoundTarget'
        The upper compound target.
    """
    # --- calculation with uneven weights
    if target_HIGH.has_uneven_weights:
        d_low = target_LOW.get_distance(vkey)  # float
        ds_high = target_HIGH.get_all_distances_for_vkey(vkey)  # list of floats (# number_of_boundaries)

        if target_HIGH.number_of_boundaries > 1:
            weights_remapped = [remap_unbound(weight, 0, weight_max, 0, 1)
                                for weight_max in target_HIGH.weight_max_per_cluster]
            weights = weights_remapped
        else:
            weights = [weight]

        distances = [(weight - 1) * d_low + weight * d_high for d_high, weight in zip(ds_high, weights)]

        if target_HIGH.has_blend_union:
            return blend_union_list(distances, target_HIGH.blend_radius)
        else:
            return min(distances)

    # --- simple calculation (without uneven weights)
    else:
        d_low = target_LOW.get_distance(vkey)
        d_high = target_HIGH.get_distance(vkey)
        return (d_low * (1 - weight)) - (d_high * weight)


if __name__ == "__main__":
    pass



import numpy as np
from compas_slicer.slicers import BaseSlicer
import logging
import progressbar
from compas_slicer.parameters import get_param
from compas_slicer.pre_processing import assign_interpolation_distance_to_mesh_vertices
from compas_slicer.slicers.slice_utilities import ScalarFieldContours
from compas_slicer.geometry import VerticalLayersManager

logger = logging.getLogger('logger')

__all__ = ['InterpolationSlicer']


class InterpolationSlicer(BaseSlicer):
    """
    Generates non-planar contours that interpolate user-defined boundaries.

    Attributes
    ----------
    mesh: :class: 'compas.datastructures.Mesh'
        Input mesh, it must be a triangular mesh (i.e. no quads or n-gons allowed)
        Note that the topology of the mesh matters, irregular tesselation can lead to undesired results.
        We recommend to 1)re-topologize, 2) triangulate, and 3) weld your mesh in advance.
    preprocessor: :class: 'compas_slicer.pre_processing.InterpolationSlicingPreprocessor'
    parameters: dict
    """

    def __init__(self, mesh, preprocessor=None, parameters=None):
        logger.info('InterpolationSlicer')
        BaseSlicer.__init__(self, mesh)

        if preprocessor:  # make sure the mesh of the preprocessor and the mesh of the slicer match
            assert len(list(mesh.vertices())) == len(list(preprocessor.mesh.vertices()))

        self.parameters = parameters if parameters else {}
        self.preprocessor = preprocessor
        self.n_multiplier = 1.0

    def generate_paths(self):
        """ Generates curved paths. """
        assert self.preprocessor, 'You need to provide a pre-processor in order to generate paths.'

        avg_layer_height = get_param(self.parameters, key='avg_layer_height', defaults_type='layers')
        n = find_no_of_isocurves(self.preprocessor.target_LOW, self.preprocessor.target_HIGH, avg_layer_height)
        params_list = get_interpolation_parameters_list(n)
        logger.info('%d paths will be generated' % n)

        max_dist = get_param(self.parameters, key='vertical_layers_max_centroid_dist', defaults_type='layers')
        vertical_layers_manager = VerticalLayersManager(max_dist)

        # create paths + layers
        with progressbar.ProgressBar(max_value=len(params_list)) as bar:
            for i, param in enumerate(params_list):
                assign_interpolation_distance_to_mesh_vertices(self.mesh, param, self.preprocessor.target_LOW,
                                                               self.preprocessor.target_HIGH)
                contours = ScalarFieldContours(self.mesh)
                contours.compute()
                contours.add_to_vertical_layers_manager(vertical_layers_manager)

                bar.update(i)  # advance progress bar

        self.layers = vertical_layers_manager.layers


def find_no_of_isocurves(target_0, target_1, avg_layer_height=1.1):
    """ Returns the average number of isocurves that can cover the get_distance from target_0 to target_1. """
    extreme_ds0 = target_0.get_extreme_distances_from_other_target(target_1)
    extreme_ds1 = target_1.get_extreme_distances_from_other_target(target_0)
    number_of_curves = max(max(extreme_ds0), max(extreme_ds1)) / avg_layer_height
    return max(1, int(number_of_curves))


def get_interpolation_parameters_list(number_of_curves):
    """ Returns a list of #number_of_curves floats from 0.001 to 0.997. """
    t_list = [0.001]
    a = list(np.arange(number_of_curves + 1) / (number_of_curves + 1))
    a.pop(0)
    t_list.extend(a)
    # t_list.append(0.997)
    return t_list


if __name__ == "__main__":
    pass



import logging
from compas.geometry import distance_point_point
from compas.geometry import Vector
import compas_slicer

logger = logging.getLogger('logger')

__all__ = ['seams_smooth']


def seams_smooth(slicer, smooth_distance):
    """Smooths the seams (transition between layers)
    by removing points within a certain distance.

    Parameters
    ----------
    slicer: :class:`compas_slicer.slicers.BaseSlicer`
        An instance of one of the compas_slicer.slicers classes.
    smooth_distance: float
        Distance (in mm) to perform smoothing
    """

    logger.info("Smoothing seams with a distance of %i mm" % smooth_distance)

    for i, layer in enumerate(slicer.layers):
        if len(layer.paths) == 1 or isinstance(layer, compas_slicer.geometry.VerticalLayer):
            for path in layer.paths:
                pt0 = path.points[0]
                # only points in the first half of a path should be evaluated
                half_of_path = path.points[:int(len(path.points)/2)]
                for point in half_of_path:
                    if distance_point_point(pt0, point) < smooth_distance:
                        # remove points if within smooth_distance
                        path.points.pop(0)
                    else:
                        # create new point at a distance of the
                        # 'smooth_distance' from the first point,
                        # so that all seams are of equal length
                        vect = Vector.from_start_end(pt0, point)
                        vect.unitize()
                        new_pt = pt0 + (vect * smooth_distance)
                        path.points.insert(0, new_pt)
                        break
        else:
            logger.warning("Smooth seams only works for layers consisting out of a single path, or for vertical layers."
                           "\nPaths were not changed, seam smoothing skipped for layer %i" % i)


if __name__ == "__main__":
    pass



import logging
from compas_slicer.print_organization import BasePrintOrganizer
import compas_slicer.utilities as utils
from compas_slicer.geometry import PrintPoint
from compas.geometry import Vector
import progressbar
import compas_slicer

logger = logging.getLogger('logger')

__all__ = ['PlanarPrintOrganizer']


class PlanarPrintOrganizer(BasePrintOrganizer):
    """
    Organizing the printing process for the realization of planar contours.

    Attributes
    ----------
    slicer: :class:`compas_slicer.slicers.PlanarSlicer`
        An instance of the compas_slicer.slicers.PlanarSlicer.
    """

    def __init__(self, slicer):
        assert isinstance(slicer, compas_slicer.slicers.PlanarSlicer), 'Please provide a PlanarSlicer'
        BasePrintOrganizer.__init__(self, slicer)

    def __repr__(self):
        return "<PlanarPrintOrganizer with %i layers>" % len(self.slicer.layers)

    def create_printpoints(self):
        """ Create the print points of the fabrication process """
        count = 0
        logger.info('Creating print points ...')
        with progressbar.ProgressBar(max_value=self.slicer.number_of_points) as bar:

            # fast method for getting the closest mesh normals to all the printpoints
            all_pts = [pt for layer in self.slicer.layers for path in layer.paths for pt in path.points]
            closest_fks, projected_pts = utils.pull_pts_to_mesh_faces(self.slicer.mesh, all_pts)
            normals = [Vector(*self.slicer.mesh.face_normal(fkey)) for fkey in closest_fks]

            for i, layer in enumerate(self.slicer.layers):
                self.printpoints_dict['layer_%d' % i] = {}

                for j, path in enumerate(layer.paths):
                    self.printpoints_dict['layer_%d' % i]['path_%d' % j] = []

                    for k, point in enumerate(path.points):

                        n = normals[count]
                        if layer.is_raft or layer.is_brim:  # then project normal on the xy plane
                            n = Vector(n[0], n[1], 0.0)

                        printpoint = PrintPoint(pt=point, layer_height=self.slicer.layer_height, mesh_normal=n)

                        self.printpoints_dict['layer_%d' % i]['path_%d' % j].append(printpoint)
                        bar.update(count)
                        count += 1


if __name__ == "__main__":
    pass



import os
import json
import logging
from compas.geometry import Point, distance_point_point_sqrd, normalize_vector
from compas.geometry import Vector, length_vector, closest_point_in_cloud, closest_point_on_plane
import matplotlib.pyplot as plt
import networkx as nx
import numpy as np
import scipy

logger = logging.getLogger('logger')

__all__ = ['remap',
           'remap_unbound',
           'get_output_directory',
           'save_to_json',
           'load_from_json',
           'is_jsonable',
           'get_jsonable_attributes',
           'save_to_text_file',
           'flattened_list_of_dictionary',
           'interrupt',
           'point_list_to_dict',
           'point_list_from_dict',
           'get_closest_mesh_vkey_to_pt',
           'get_mesh_cotmatrix_igl',
           'get_mesh_cotans_igl',
           'get_closest_pt_index',
           'get_closest_pt',
           'pull_pts_to_mesh_faces',
           'plot_networkx_graph',
           'get_mesh_vertex_coords_with_attribute',
           'get_dict_key_from_value',
           'find_next_printpoint',
           'find_previous_printpoint',
           'smooth_vectors',
           'get_normal_of_path_on_xy_plane',
           'get_all_files_with_name']


def remap(input_val, in_from, in_to, out_from, out_to):
    """ Bounded remap. """
    if input_val <= in_from:
        return out_from
    elif input_val >= in_to:
        return out_to
    else:
        return remap_unbound(input_val, in_from, in_to, out_from, out_to)


def remap_unbound(input_val, in_from, in_to, out_from, out_to):
    """
    Remaps input_val from source domain to target domain.
    No clamping is performed, the result can be outside of the target domain
    if the input is outside of the source domain.
    """
    out_range = out_to - out_from
    in_range = in_to - in_from
    in_val = input_val - in_from
    val = (float(in_val) / in_range) * out_range
    out_val = out_from + val
    return out_val


def get_output_directory(path):
    """
    Checks if a directory with the name 'output' exists in the path. If not it creates it.

    Parameters
    ----------
    path: str
        The path where the 'output' directory will be created

    Returns
    ----------
    str
        The path to the new (or already existing) 'output' directory
    """
    output_dir = os.path.join(path, 'output')
    if not os.path.exists(output_dir):
        os.mkdir(output_dir)
    return output_dir


def get_closest_pt_index(pt, pts):
    """
    Finds the index of the closest point of 'pt' in the point cloud 'pts'.

    Parameters
    ----------
    pt: compas.geometry.Point3d
    pts: list, compas.geometry.Point3d

    Returns
    ----------
    int
        The index of the closest point
    """
    ci = closest_point_in_cloud(point=pt, cloud=pts)[2]
    # distances = [distance_point_point_sqrd(p, pt) for p in pts]
    # ci = distances.index(min(distances))
    return ci


def get_closest_pt(pt, pts):
    """
     Finds the closest point of 'pt' in the point cloud 'pts'.

    Parameters
    ----------
    pt: :class: 'compas.geometry.Point'
    pts: list, :class: 'compas.geometry.Point3d'

    Returns
    ----------
    compas.geometry.Point3d
        The closest point
    """
    ci = closest_point_in_cloud(point=pt, cloud=pts)[2]
    return pts[ci]


def pull_pts_to_mesh_faces(mesh, points):
    """
    Very fast method for projecting a list of points on a mesh, and finding their closest face keys.

    Parameters
    ----------
    mesh: :class: compas.datastructures.Mesh
    points: list, compas.geometry.Point

    Returns
    -------
    closest_fks: a list of the closest face keys
    projected_pts: a list of the projected points on the mesh
    """
    points = np.array(points, dtype=np.float64).reshape((-1, 3))
    fi_fk = {index: fkey for index, fkey in enumerate(mesh.faces())}
    f_centroids = np.array([mesh.face_centroid(fkey) for fkey in mesh.faces()], dtype=np.float64)
    closest_fis = np.argmin(scipy.spatial.distance_matrix(points, f_centroids), axis=1)
    closest_fks = [fi_fk[fi] for fi in closest_fis]
    projected_pts = [closest_point_on_plane(point, mesh.face_plane(fi)) for point, fi in zip(points, closest_fis)]
    return closest_fks, projected_pts


def smooth_vectors(vectors, strength, iterations):
    """
    Smooths the vector iteratively, with the given number of iterations and strength per iteration

    Parameters
    ----------
    vectors: list, :class: 'compas.geometry.Vector'
    strength: float
    iterations: int

    Returns
    ----------
    list, :class: 'compas.geometry.Vector3d'
        The smoothened vectors
    """

    for _ in range(iterations):
        for i, n in enumerate(vectors):
            if 0 < i < len(vectors) - 1:
                neighbors_average = (vectors[i - 1] + vectors[i + 1]) * 0.5
            else:
                neighbors_average = n
            vectors[i] = n * (1 - strength) + neighbors_average * strength
    return vectors


#######################################
#  json

def save_to_json(data, filepath, name):
    """
    Save the provided data to json on the filepath, with the given name

    Parameters
    ----------
    data: dict_or_list
    filepath: str
    name: str
    """

    filename = os.path.join(filepath, name)
    logger.info("Saving to json: " + filename)
    with open(filename, 'w') as f:
        f.write(json.dumps(data, indent=3, sort_keys=True))


def load_from_json(filepath, name):
    """
    Loads json from the filepath

    Parameters
    ----------
    filepath: str
    name: str
    """

    filename = os.path.join(filepath, name)
    with open(filename, 'r') as f:
        data = json.load(f)
    logger.info("Loaded json: " + filename)
    return data


def is_jsonable(x):
    """ Returns True if x can be json-serialized, False otherwise. """
    try:
        json.dumps(x)
        return True
    except TypeError:
        return False


def get_jsonable_attributes(attributes_dict):
    jsonable_attr = {}
    for attr_key in attributes_dict:
        attr = attributes_dict[attr_key]
        if is_jsonable(attr):
            jsonable_attr[attr_key] = attr
        else:
            if isinstance(attr, np.ndarray):
                jsonable_attr[attr_key] = list(attr)
            else:
                jsonable_attr[attr_key] = 'non serializable attribute'

    return jsonable_attr


#######################################
#  text file

def save_to_text_file(data, filepath, name):
    """
    Save the provided text on the filepath, with the given name

    Parameters
    ----------
    data: str
    filepath: str
    name: str
    """

    filename = os.path.join(filepath, name)
    logger.info("Saving to text file: " + filename)
    with open(filename, 'w') as f:
        f.write(data)


#######################################
#  mesh utils

def check_triangular_mesh(mesh):
    """
    Checks if the mesh is triangular. If not, then it raises an error

    Parameters
    ----------
    mesh: :class: 'compas.datastructures.Mesh'
    """

    for f_key in mesh.faces():
        vs = mesh.face_vertices(f_key)
        if len(vs) != 3:
            raise TypeError("Found a quad at face key: " + str(f_key) + " ,number of face vertices:" + str(
                len(vs)) + ". \nOnly triangular meshes supported.")


def get_closest_mesh_vkey_to_pt(mesh, pt):
    """
    Finds the vertex key that is the closest to the point.

    Parameters
    ----------
    mesh: :class: 'compas.datastructures.Mesh'
    pt: :class: 'compas.geometry.Point'

    Returns
    ----------
    int
        the closest vertex key
    """
    # cloud = [Point(data['x'], data['y'], data['z']) for v_key, data in mesh.vertices(data=True)]
    # closest_index = compas.geometry.closest_point_in_cloud(pt, cloud)[2]
    vertex_tupples = [(v_key, Point(data['x'], data['y'], data['z'])) for v_key, data in mesh.vertices(data=True)]
    vertex_tupples = sorted(vertex_tupples, key=lambda v_tupple: distance_point_point_sqrd(pt, v_tupple[1]))
    closest_vkey = vertex_tupples[0][0]
    return closest_vkey


def get_closest_mesh_normal_to_pt(mesh, pt):
    """
    Finds the closest vertex normal to the point.
    Parameters
    ----------
    mesh: :class: 'compas.datastructures.Mesh'
    pt: :class: 'compas.geometry.Point'
    Returns
    ----------
    :class: 'compas.geometry.Vector'
        The closest normal of the mesh.
    """

    closest_vkey = get_closest_mesh_vkey_to_pt(mesh, pt)
    v = mesh.vertex_normal(closest_vkey)
    return Vector(v[0], v[1], v[2])


def get_mesh_vertex_coords_with_attribute(mesh, attr, value):
    """
    Finds the coordinates of all the vertices that have an attribute with key=attr that equals the value.

    Parameters
    ----------
    mesh: :class: 'compas.datastructures.Mesh'
    attr: str
    value: anything that can be stored into a dictionary

    Returns
    ----------
    list, :class: 'compas.geometry.Point'
        the closest vertex key
    """

    pts = []
    for vkey, data in mesh.vertices(data=True):
        if data[attr] == value:
            pts.append(Point(*mesh.vertex_coordinates(vkey)))
    return pts


def get_normal_of_path_on_xy_plane(k, point, path, mesh):
    """
    Finds the normal of the curve that lies on the xy plane at the point with index k

    Parameters
    ----------
    k: int, index of the point
    point: :class: 'compas.geometry.Point'
    path: :class: 'compas_slicer.geometry.Path'
    mesh: :class: 'compas.datastructures.Mesh'

    Returns
    ----------
    :class: 'compas.geometry.Vector'
    """

    # find mesh normal is not really needed in the 2D case of planar slicer
    # instead we only need the normal of the curve based on the neighboring pts
    if (0 < k < len(path.points) - 1) or path.is_closed:
        prev_pt = path.points[k - 1]
        next_pt = path.points[(k + 1) % len(path.points)]
        v1 = np.array(normalize_vector(Vector.from_start_end(prev_pt, point)))
        v2 = np.array(normalize_vector(Vector.from_start_end(point, next_pt)))
        v = (v1 + v2) * 0.5
        normal = [-v[1], v[0], v[2]]  # rotate 90 degrees COUNTER-clockwise on the xy plane

    else:
        if k == 0:
            next_pt = path.points[k + 1]
            v = normalize_vector(Vector.from_start_end(point, next_pt))
            normal = [-v[1], v[0], v[2]]  # rotate 90 degrees COUNTER-clockwise on the xy plane
        else:  # k == len(path.points)-1:
            prev_pt = path.points[k - 1]
            v = normalize_vector(Vector.from_start_end(point, prev_pt))
            normal = [v[1], -v[0], v[2]]  # rotate 90 degrees clockwise on the xy plane

    if length_vector(normal) == 0:
        # When the neighboring elements happen to cancel out, then search for the true normal,
        # and project it on the xy plane for consistency
        normal = get_closest_mesh_normal_to_pt(mesh, point)
        normal = [normal[0], normal[1], 0]

    normal = normalize_vector(normal)
    normal = Vector(*list(normal))
    return normal


#######################################
# igl utils

def get_mesh_cotmatrix_igl(mesh, fix_boundaries=True):
    """
    Gets the laplace operator of the mesh

    Parameters
    ----------
    mesh: :class: 'compas.datastructures.Mesh'
    fix_boundaries : bool

    Returns
    ----------
    :class: 'scipy.sparse.csr_matrix'
        sparse matrix (dimensions: #V x #V), laplace operator, each row i corresponding to v(i, :)
    """
    import igl
    v, f = mesh.to_vertices_and_faces()
    C = igl.cotmatrix(np.array(v), np.array(f))

    if fix_boundaries:
        # fix boundaries by putting the corresponding columns of the sparse matrix to 0
        C_dense = C.toarray()
        for i, (vkey, data) in enumerate(mesh.vertices(data=True)):
            if data['boundary'] > 0:
                C_dense[i][:] = np.zeros(len(v))
        C = scipy.sparse.csr_matrix(C_dense)
    return C


def get_mesh_cotans_igl(mesh):
    """
    Gets the cotangent entries of the mesh


    Parameters
    ----------
    mesh: :class: 'compas.datastructures.Mesh'

    Returns
    ----------
    :class: 'np.array'
        Dimensions: F by 3 list of 1/2*cotangents corresponding angles
    """
    import igl
    v, f = mesh.to_vertices_and_faces()
    return igl.cotmatrix_entries(np.array(v), np.array(f))


#######################################
#  networkx graph

def plot_networkx_graph(G):
    """
    Plots the graph G

    Parameters
    ----------
    G: networkx.Graph
    """

    plt.subplot(121)
    nx.draw(G, with_labels=True, font_weight='bold', node_color=range(len(list(G.nodes()))))
    plt.show()


#######################################
#  dict utils

def point_list_to_dict(pts_list):
    """
    Turns a list of compas.geometry.Point into a dictionary, so that it can be saved to Json. Works identically for
    3D vectors.

    Parameters
    ----------
    pts_list: list, :class:`compas.geometry.Point` / :class:`compas.geometry.Vector`

    Returns
    ----------
    dict: The dictionary of pts in the form { key=index : [x,y,z] }
    """
    data = {}
    for i in range(len(pts_list)):
        data[i] = list(pts_list[i])
    return data


def point_list_from_dict(data):
    """
    Turns a dictionary of pts to a list of Compas.geometry.Point. Works identically for 3D vectors.

    Parameters
    ----------
    dict: The dictionary of pts in the form { key=index : [x,y,z] }

    Returns
    ----------
    2D list,  [[x1, y1, z1], ... , [xn, yn, zn]]
    """
    return [[data[i][0], data[i][1], data[i][2]] for i in data]


#  --- Flattened list of dictionary
def flattened_list_of_dictionary(dictionary):
    """
    Turns the dictionary into a flat list

    Parameters
    ----------
    dictionary: dict

    Returns
    ----------
    list
    """
    flattened_list = []
    for key in dictionary:
        [flattened_list.append(item) for item in dictionary[key]]
    return flattened_list


def get_dict_key_from_value(dictionary, val):
    """
    Return the key of a dictionary that stores the val

    Parameters
    ----------
    dictionary: dict
    val: anything that can be stored in a dictionary
    """

    for key in dictionary:
        value = dictionary[key]
        if val == value:
            return key
    return "key doesn't exist"


def find_next_printpoint(pp_dict, i, j, k):
    """
    Returns the next printpoint from the current printpoint if it exists, otherwise returns None.
    """
    next_ppt = None
    layer_key, path_key = 'layer_%d' % i, 'path_%d' % j
    if k < len(pp_dict[layer_key][path_key]) - 1:  # If there are more ppts in the current path, then take the next ppt
        next_ppt = pp_dict[layer_key][path_key][k + 1]
    else:
        if j < len(pp_dict[layer_key]) - 1:  # Otherwise take the next path if there are more paths in the current layer
            next_ppt = pp_dict[layer_key]['path_%d' % (j + 1)][0]
        else:
            if i < len(pp_dict) - 1:  # Otherwise take the next layer if there are more layers in the current slicer
                next_ppt = pp_dict['layer_%d' % (i + 1)]['path_0'][0]
    return next_ppt


def find_previous_printpoint(pp_dict, layer_key, path_key, i, j, k):
    """
    Returns the previous printpoint from the current printpoint if it exists, otherwise returns None.
    """
    prev_ppt = None
    if k > 0:  # If not the first point in a path, take the previous point in the path
        prev_ppt = pp_dict[layer_key][path_key][k - 1]
    else:
        if j > 0:  # Otherwise take the last point of the previous path, if there are more paths in the current layer
            prev_ppt = pp_dict[layer_key]['path_%d' % (j - 1)][-1]
        else:
            if i > 0:  # Otherwise take the last path of the previous layer if there are more layers in the current slicer
                last_path_key = len(pp_dict[layer_key]) - 1
                prev_ppt = pp_dict['layer_%d' % (i - 1)]['path_%d' % (last_path_key)][-1]
    return prev_ppt


#######################################
#  control flow

def interrupt():
    """
    Interrupts the flow of the code while it is running.
    It asks for the user to press a enter to continue or abort.
    """

    value = input("Press enter to continue, Press 1 to abort ")
    print("")
    if isinstance(value, str):
        if value == '1':
            raise ValueError("Aborted")


#######################################
#  load all files with name

def get_all_files_with_name(startswith, endswith, DATA_PATH):
    """
    Finds all the filenames in the DATA_PATH that start and end with the provided strings

    Parameters
    ----------
    startswith: str
    endswith: str
    DATA_PATH: str

    Returns
    ----------
    list, str
        All the filenames
    """

    files = []
    for file in os.listdir(DATA_PATH):
        if file.startswith(startswith) and file.endswith(endswith):
            files.append(file)
    print('')
    logger.info('Reloading : ' + str(files))
    return files


if __name__ == "__main__":
    pass



from compas_slicer.geometry import VerticalLayersManager
import logging

logger = logging.getLogger('logger')

__all__ = ['sort_into_vertical_layers']


def sort_into_vertical_layers(slicer, dist_threshold=25.0, max_paths_per_layer=None):
    """Sorts the paths from horizontal layers into Vertical Layers.

    Vertical Layers are layers at different heights that are grouped together by proximity
    of their center points. Can be useful for reducing travel time in a robotic printing
    process.

    Parameters
    ----------
    slicer: :class:`compas_slicer.slicers.BaseSlicer`
        An instance of one of the compas_slicer.slicers classes.
    dist_threshold: float
        The maximum get_distance that the centroids of two successive paths can have to belong in the same VerticalLayer
        Recommended value, slightly bigger than the layer height
    max_paths_per_layer: int
        Maximum number of layers that a vertical layer can consist of.
        If None, then the vertical layer has an unlimited number of layers.
    """
    logger.info("Sorting into Vertical Layers")

    vertical_layers_manager = VerticalLayersManager(dist_threshold, max_paths_per_layer)

    for layer in slicer.layers:
        for path in layer.paths:
            vertical_layers_manager.add(path)

    slicer.layers = vertical_layers_manager.layers
    logger.info("Number of vertical_layers: %d" % len(slicer.layers))


if __name__ == "__main__":
    pass



import logging
from compas.geometry import normalize_vector, subtract_vectors, dot_vectors
from collections import deque

logger = logging.getLogger('logger')

__all__ = ['unify_paths_orientation']


def unify_paths_orientation(slicer):
    """
    Unifies the orientation of paths that are closed.

    Parameters
    ----------
    slicer: :class:`compas_slicer.slicers.BaseSlicer`
        An instance of one of the compas_slicer.slicers classes.
    """

    for i, layer in enumerate(slicer.layers):
        for j, path in enumerate(layer.paths):
            # find reference points for each path, if possible
            if path.is_closed:
                reference_points = None
                if j > 0:
                    reference_points = layer.paths[0].points
                elif i > 0 and j == 0:
                    reference_points = slicer.layers[i - 1].paths[0].points

                if reference_points:  # then reorient current pts based on reference
                    path.points = make_pts_have_same_direction_as_pts_reference(path.points,
                                                                                reference_points)


def make_pts_have_same_direction_as_pts_reference(pts, reference_points):
    """
    Check if new curve has same direction as prev curve, otherwise reverse.

    Parameters
    ----------
    pts: list, :class: 'compas.geometry.Point'. The list of points whose direction we are fixing.
    reference_points: list, :class: 'compas.geometry.Point'. [p1, p2] Two reference points.
    """
    if len(pts) > 2 and len(reference_points) > 2:
        v1 = normalize_vector(subtract_vectors(pts[0], pts[2]))
        v2 = normalize_vector(subtract_vectors(reference_points[0], reference_points[2]))
    else:
        v1 = normalize_vector(subtract_vectors(pts[0], pts[1]))
        v2 = normalize_vector(subtract_vectors(reference_points[0], reference_points[1]))

    if dot_vectors(v1, v2) < 0:
        items = deque(reversed(pts))
        items.rotate(1)  # bring last point again in the front
        pts = list(items)
    return pts


if __name__ == "__main__":
    pass



from compas_slicer.slicers.slice_utilities import ContoursBase
from compas.geometry import Vector, add_vectors, scale_vector

__all__ = ['ScalarFieldContours']


class ScalarFieldContours(ContoursBase):
    """
    Finds the iso-contours of the function f(x) = vertex_data['scalar_field']
    on the mesh.

    Attributes
    ----------
    mesh: :class: 'compas.datastructures.Mesh'
    """
    def __init__(self, mesh):
        ContoursBase.__init__(self, mesh)  # initialize from parent class

    def edge_is_intersected(self, u, v):
        """ Returns True if the edge u,v has a zero-crossing, False otherwise. """
        d1 = self.mesh.vertex[u]['scalar_field']
        d2 = self.mesh.vertex[v]['scalar_field']
        if (d1 > 0 and d2 > 0) or (d1 < 0 and d2 < 0):
            return False
        else:
            return True

    def find_zero_crossing_data(self, u, v):
        """ Finds the position of the zero-crossing on the edge u,v. """
        dist_a, dist_b = self.mesh.vertex[u]['scalar_field'], self.mesh.vertex[v]['scalar_field']
        if abs(dist_a) + abs(dist_b) > 0:
            v_coords_a, v_coords_b = self.mesh.vertex_coordinates(u), self.mesh.vertex_coordinates(v)
            vec = Vector.from_start_end(v_coords_a, v_coords_b)
            vec = scale_vector(vec, abs(dist_a) / (abs(dist_a) + abs(dist_b)))
            pt = add_vectors(v_coords_a, vec)
            return pt


if __name__ == "__main__":
    pass



from compas.geometry import Vector, dot_vectors
from compas_slicer.utilities import remap, remap_unbound
import logging

logger = logging.getLogger('logger')

__all__ = ['set_linear_velocity_constant',
           'set_linear_velocity_per_layer',
           'set_linear_velocity_by_range',
           'set_linear_velocity_by_overhang']


def set_linear_velocity_constant(print_organizer, v=25.0):
    """ Sets the linear velocity parameter of the printpoints depending on the selected type.

    Parameters
    ----------
    print_organizer: :class:`compas_slicer.print_organization.BasePrintOrganizer`
    v:  float. Velocity value (in mm/s) to set for printpoints. Defaults to 25 mm/s. """
    logger.info("Setting constant linear velocity")
    for printpoint in print_organizer.printpoints_iterator():
        printpoint.velocity = v


def set_linear_velocity_per_layer(print_organizer, per_layer_velocities):
    """ Sets the linear velocity parameter of the printpoints depending on the selected type.

    Parameters
    ----------
    print_organizer: :class:`compas_slicer.print_organization.BasePrintOrganizer`
    per_layer_velocities: list
        A list of velocities (floats) with equal length to the number of layers.
    """
    logger.info("Setting per-layer linear velocity")
    assert len(per_layer_velocities) == print_organizer.number_of_layers, 'Wrong number of velocity values. You need \
        to provide one velocity value per layer, on the "per_layer_velocities" list.'
    for printpoint, i, j, k in print_organizer.printpoints_indices_iterator():
        printpoint.velocity = per_layer_velocities[i]


def set_linear_velocity_by_range(print_organizer, param_func, parameter_range, velocity_range,
                                 bound_remapping=True):
    """
    Sets the linear velocity parameter of the printpoints depending on the selected type.

    Parameters
    ----------
    print_organizer: :class:`compas_slicer.print_organization.BasePrintOrganizer`
    param_func: function that takes as argument a :class: 'compas_slicer.geometry.Printpoint': get_param_func(pp)
        and returns the parameter value that will be used for the remapping
    parameter_range: tuple
        An example of a parameter that can be used is the overhang angle, or the layer height.
    velocity_range: tuple
        The range of velocities where the parameter will be remapped
    bound_remapping: bool
        If True, the remapping is bound in the domain velocity_range, else it is unbound.
    """
    logger.info("Setting linear velocity based on parameter range")
    for printpoint in print_organizer.printpoints_iterator():
        param = param_func(printpoint)
        assert param, 'The param_func does not return any value for calculating the velocity range.'
        if bound_remapping:
            v = remap(param, parameter_range[0], parameter_range[1], velocity_range[0], velocity_range[1])
        else:
            v = remap_unbound(param, parameter_range[0], parameter_range[1], velocity_range[0], velocity_range[1])
        printpoint.velocity = v


def set_linear_velocity_by_overhang(print_organizer, overhang_range, velocity_range, bound_remapping=True):
    """
    An example function for how to use the 'set_linear_velocity_by_range'. In this case the parameter that controls the
    velocity is the overhang, measured as a dot product with the horizontal direction.

    Parameters
    ----------
    print_organizer: :class:`compas_slicer.print_organization.BasePrintOrganizer`
    overhang_range: tuple:
        should be within [0.0, 1.0]. For example a reasonable value would be [0.0, 0.5], that would
        be remapping overhangs up to 45 degrees
    velocity_range: tuple
    bound_remapping: bool
    """
    def param_func(ppt): return dot_vectors(ppt.mesh_normal, Vector(0.0, 0.0, 1.0))
    # returns values from 0.0 (no overhang) to 1.0 (horizontal overhang)
    set_linear_velocity_by_range(print_organizer, param_func, overhang_range, velocity_range, bound_remapping)


if __name__ == "__main__":
    pass



import compas_slicer
import logging
from compas.geometry import Polyline, distance_point_point
from compas.utilities import pairwise
import numpy as np
from abc import abstractmethod

logger = logging.getLogger('logger')

__all__ = ['BasePrintOrganizer']


class BasePrintOrganizer(object):
    """
    Base class for organizing the printing process.
    This class is meant to be extended for the implementation of the various print organizers.
    Do not use this class directly in your python code. Instead use PlanarPrintOrganizer or InterpolationPrintOrganizer.

    Attributes
    ----------
    slicer: :class:`compas_slicer.slicers.PlanarSlicer`
        An instance of the compas_slicer.slicers.PlanarSlicer.
    """

    def __init__(self, slicer):
        assert isinstance(slicer, compas_slicer.slicers.BaseSlicer)  # check input
        logger.info('Print Organizer')
        self.slicer = slicer
        self.printpoints_dict = {}

    def __repr__(self):
        return "<BasePrintOrganizer>"

    ######################
    # Abstract methods
    ######################

    @abstractmethod
    def create_printpoints(self):
        """To be implemented by the inheriting classes"""
        pass

    ######################
    # Iterators
    ######################
    def printpoints_iterator(self):
        """Iterate over the printpoints of the print organizer.
        Yields
        ------
        printpoint: :class: 'compas_slicer.geometry.Printpoint'
        """
        assert len(self.printpoints_dict) > 0, 'No printpoints have been created.'
        for layer_key in self.printpoints_dict:
            for path_key in self.printpoints_dict[layer_key]:
                for printpoint in self.printpoints_dict[layer_key][path_key]:
                    yield printpoint

    def printpoints_indices_iterator(self):
        """ Iterate over the printpoints of the print organizer.
        Yields
        ------
        printpoint: :class: 'compas_slicer.geometry.Printpoint'
        i: int, layer index. To get the layer key use: layer_key = 'layer_%d' % i
        j: int, path index. To get the path key use: path_key = 'path_%d' % j
        k: int, printpoint index
        """
        assert len(self.printpoints_dict) > 0, 'No printpoints have been created.'
        for i, layer_key in enumerate(self.printpoints_dict):
            for j, path_key in enumerate(self.printpoints_dict[layer_key]):
                for k, printpoint in enumerate(self.printpoints_dict[layer_key][path_key]):
                    yield printpoint, i, j, k

    ######################
    # Properties
    ######################

    @property
    def number_of_printpoints(self):
        """int: Total number of points in the PrintOrganizer."""
        total_number_of_pts = 0
        for layer_key in self.printpoints_dict:
            for path_key in self.printpoints_dict[layer_key]:
                for _ in self.printpoints_dict[layer_key][path_key]:
                    total_number_of_pts += 1
        return total_number_of_pts

    @property
    def number_of_paths(self):
        total_number_of_paths = 0
        for layer_key in self.printpoints_dict:
            for _ in self.printpoints_dict[layer_key]:
                total_number_of_paths += 1
        return total_number_of_paths

    @property
    def number_of_layers(self):
        """int: Number of layers in the PrintOrganizer."""
        return len(self.printpoints_dict)

    @property
    def total_length_of_paths(self):
        """ Returns the total length of all paths. Does not consider extruder toggle. """
        total_length = 0
        for layer_key in self.printpoints_dict:
            for path_key in self.printpoints_dict[layer_key]:
                for prev, curr in pairwise(self.printpoints_dict[layer_key][path_key]):
                    length = distance_point_point(prev.pt, curr.pt)
                    total_length += length
        return total_length

    @property
    def total_print_time(self):
        """ If the print speed is defined, it returns the total time of the print, else returns None"""
        if self.printpoints_dict['layer_0']['path_0'][0].velocity is not None:  # assume that all ppts are set or none
            total_time = 0
            for layer_key in self.printpoints_dict:
                for path_key in self.printpoints_dict[layer_key]:
                    for prev, curr in pairwise(self.printpoints_dict[layer_key][path_key]):
                        length = distance_point_point(prev.pt, curr.pt)
                        total_time += length / curr.velocity
            return total_time

    def number_of_paths_on_layer(self, layer_index):
        """int: Number of paths within a Layer of the PrintOrganizer."""
        return len(self.printpoints_dict['layer_%d' % layer_index])

    ######################
    # Utils
    ######################

    def remove_duplicate_points_in_path(self, layer_key, path_key, tolerance=0.0001):
        """Remove subsequent points that are within a certain tolerance.

        Parameters
        ----------
        layer_key: str
            They key of the layer to remove points from.
        path_key: str
            The key of the path to remove points from.
        tolerance: float, optional
            Distance between points to remove. Defaults to 0.0001.
        """

        dup_index = []
        # find duplicates
        duplicate_ppts = []
        for i, printpoint in enumerate(self.printpoints_dict[layer_key][path_key]):
            if i < len(self.printpoints_dict[layer_key][path_key]) - 1:
                next_ppt = self.printpoints_dict[layer_key][path_key][i + 1]
                if np.linalg.norm(np.array(printpoint.pt) - np.array(next_ppt.pt)) < tolerance:
                    dup_index.append(i)
                    duplicate_ppts.append(printpoint)

        # warn user
        if len(duplicate_ppts) > 0:
            logger.warning(
                'Attention! %d Duplicate printpoint(s) ' % len(duplicate_ppts)
                + 'on ' + layer_key + ', ' + path_key + ', indices: ' + str(dup_index) + '. They will be removed.')

        # remove duplicates
        if len(duplicate_ppts) > 0:
            for ppt in duplicate_ppts:
                self.printpoints_dict[layer_key][path_key].remove(ppt)

    def get_printpoint_neighboring_items(self, layer_key, path_key, i):
        """
        layer_key: str
            They key of the layer the current printpoint belongs to.
        path_key: str
            They key of the path the current printpoint belongs to.
        i: int
            The index of the current printpoint.

        Returns
        ----------
        list, :class:  'compas_slicer.geometry.PrintPoint'
        """
        neighboring_items = []
        if i > 0:
            neighboring_items.append(self.printpoints_dict[layer_key][path_key][i - 1])
        else:
            neighboring_items.append(None)
        if i < len(self.printpoints_dict[layer_key][path_key]) - 1:
            neighboring_items.append(self.printpoints_dict[layer_key][path_key][i + 1])
        else:
            neighboring_items.append(None)
        return neighboring_items

    def printout_info(self):
        """Prints out information from the PrintOrganizer"""
        ppts_attributes = {}
        for key in self.printpoints_dict['layer_0']['path_0'][0].attributes:
            ppts_attributes[key] = str(type(self.printpoints_dict['layer_0']['path_0'][0].attributes[key]))

        print("\n---- PrintOrganizer Info ----")
        print("Number of layers: %d" % self.number_of_layers)
        print("Number of paths: %d" % self.number_of_paths)
        print("Number of PrintPoints: %d" % self.number_of_printpoints)
        print("PrintPoints attributes: ")
        for key in ppts_attributes:
            print('     % s : % s' % (str(key), ppts_attributes[key]))
        print("Toolpath length: %d mm" % self.total_length_of_paths)

        print_time = self.total_print_time
        if print_time:
            minutes, sec = divmod(self.total_print_time, 60)
            hour, minutes = divmod(minutes, 60)
            print("Total print time: %d hours, %d minutes, %d seconds" % (hour, minutes, sec))
        else:
            print("Print Velocity has not been assigned, thus print time is not calculated.")
        print("")

    def visualize_on_viewer(self, viewer, visualize_polyline, visualize_printpoints):
        """Visualize printpoints on the compas_viewer.

        Parameters
        ----------
        viewer: :class: 'compas_viewers.objectviewer.ObjectViewer'
        visualize_polyline: bool
        visualize_printpoints: bool
        """
        all_pts = []
        for layer_key in self.printpoints_dict:
            for path_key in self.printpoints_dict[layer_key]:
                for printpoint in self.printpoints_dict[layer_key][path_key]:
                    all_pts.append(printpoint.pt)

        if visualize_polyline:
            polyline = Polyline(all_pts)
            viewer.add(polyline, name="Polyline", settings={'color': '#ffffff'})

        if visualize_printpoints:
            [viewer.add(pt, name="Point %d" % i) for i, pt in enumerate(all_pts)]

    ######################
    # Output data
    ######################

    def output_printpoints_dict(self):
        """Creates a flattened PrintPoints as a dictionary.

        Returns
        ----------
        dict, with printpoints that can be saved as json
        """
        data = {}

        count = 0
        for layer_key in self.printpoints_dict:
            for path_key in self.printpoints_dict[layer_key]:
                self.remove_duplicate_points_in_path(layer_key, path_key)
                for printpoint in self.printpoints_dict[layer_key][path_key]:
                    data[count] = printpoint.to_data()

                    count += 1
        logger.info("Generated %d print points" % count)
        return data

    def output_gcode(self, parameters):
        """ Gets a gcode text file using the function that creates gcode
        Parameters
        ----------
        parameters: dict with gcode parameters

        Returns
        ----------
        str, gcode text file
        """
        # check print organizer: Should have horizontal layers, ideally should be planar
        # ...
        gcode = compas_slicer.print_organization.create_gcode_text(self, parameters)
        return gcode

    def get_printpoints_attribute(self, attr_name):
        """
        Returns a list of printpoint attributes that have key=attr_name.

        Parameters
        ----------
        attr_name: str

        Returns
        -------
        list of size len(ppts) with whatever type the ppts.attribute[attr_name] is.
        """
        attr_values = []
        for layer_key in self.printpoints_dict:
            for path_key in self.printpoints_dict[layer_key]:
                for ppt in self.printpoints_dict[layer_key][path_key]:
                    assert attr_name in ppt.attributes, \
                        "The attribute '%s' is not in the printpoint.attributes" % attr_name
                    attr_values.append(ppt.attributes[attr_name])
        return attr_values


if __name__ == "__main__":
    pass



from compas_slicer.pre_processing import CompoundTarget
from compas_slicer.pre_processing.gradient_evaluation import GradientEvaluation
import logging
import os
from compas.datastructures import Mesh
from compas_slicer.pre_processing.preprocessing_utils import region_split as rs, \
    topological_sorting as topo_sort
from compas_slicer.pre_processing import get_existing_cut_indices, get_vertices_that_belong_to_cuts, \
    replace_mesh_vertex_attribute
import compas_slicer.utilities as utils
from compas_slicer.parameters import get_param
from compas_slicer.pre_processing.preprocessing_utils import assign_interpolation_distance_to_mesh_vertices

logger = logging.getLogger('logger')

__all__ = ['InterpolationSlicingPreprocessor']


class InterpolationSlicingPreprocessor:
    """
    Takes care of all the pre-processing that is (or might be) needed before the interpolation slicing process.
    Not all the functionality needs to be run every time, depending on the characteristics of the inputs.

    Attributes
    ----------
    mesh: :class: 'compas.datastructures.Mesh'
    parameters: dict
    DATA_PATH: str, path to the data folder
    """

    def __init__(self, mesh, parameters, DATA_PATH):
        self.mesh = mesh
        self.parameters = parameters
        self.DATA_PATH = DATA_PATH

        self.OUTPUT_PATH = utils.get_output_directory(DATA_PATH)
        self.target_LOW = None  # :class: 'compas_slicer.pre_processing.CompoundTarget'
        self.target_HIGH = None  # :class: 'compas_slicer.pre_processing.CompoundTarget'

        self.split_meshes = []  # list , :class: 'compas.datastructures.Mesh'
        # The meshes that result from the region splitting process.

    ###########################
    # --- compound targets

    def create_compound_targets(self):
        """ Creates the target_LOW and the target_HIGH and computes the geodesic distances. """

        # --- low target
        target_blending = get_param(self.parameters, key='target_LOW_smooth_union', defaults_type='interpolation_slicing')
        smooth, r = target_blending[0], target_blending[1]
        geodesics_method = get_param(self.parameters, key='target_LOW_geodesics_method', defaults_type='interpolation_slicing')
        self.target_LOW = CompoundTarget(self.mesh, 'boundary', 1, self.DATA_PATH,
                                         has_blend_union=smooth,
                                         blend_radius=r,
                                         geodesics_method=geodesics_method)

        # --- high target
        target_blending = get_param(self.parameters, key='target_HIGH_smooth_union', defaults_type='interpolation_slicing')
        smooth, r = target_blending[0], target_blending[1]
        geodesics_method = get_param(self.parameters, key='target_HIGH_geodesics_method',
                                     defaults_type='interpolation_slicing')
        self.target_HIGH = CompoundTarget(self.mesh, 'boundary', 2, self.DATA_PATH,
                                          has_blend_union=smooth,
                                          blend_radius=r,
                                          geodesics_method=geodesics_method)

        # --- uneven boundaries of high target
        self.target_HIGH.offset = get_param(self.parameters, key='uneven_upper_targets_offset',
                                            defaults_type='interpolation_slicing')
        self.target_HIGH.compute_uneven_boundaries_weight_max(self.target_LOW)

        #  --- save intermediary get_distance outputs
        self.target_LOW.save_distances("distances_LOW.json")
        self.target_HIGH.save_distances("distances_HIGH.json")
        self.save_df_interpolation_data()

    def save_df_interpolation_data(self):
        """ Saves a dictionary with the data for visualization of the interpolation of the targets distance fields."""
        df_interpolation = {'distances_LOW': self.target_LOW.get_all_distances(),
                            'distances_HIGH_clusters': self.target_HIGH.get_all_clusters_distances_dict(),
                            't_end_per_cluster_HIGH': self.target_HIGH.weight_max_per_cluster,
                            'has_blend_union_HIGH': self.target_HIGH.has_blend_union,
                            'blend_radius': self.target_HIGH.blend_radius}
        utils.save_to_json(df_interpolation, self.OUTPUT_PATH, "df_interpolation.json")

    def targets_laplacian_smoothing(self, iterations, strength):
        """
        Smooth geodesic distances of targets. Saves again the distances to json.

        Parameters
        ----------
        iterations: int
        strength: float
        """
        self.target_LOW.laplacian_smoothing(iterations=iterations, strength=strength)
        self.target_HIGH.laplacian_smoothing(iterations=iterations, strength=strength)
        self.target_LOW.save_distances("distances_LOW.json")
        self.target_HIGH.save_distances("distances_HIGH.json")
        self.save_df_interpolation_data()

    ###########################
    # --- scalar field evaluation

    def create_gradient_evaluation(self, target_1, target_2=None, save_output=True,
                                   norm_filename='gradient_norm.json', g_filename='gradient.json'):
        """
        Creates a compas_slicer.pre_processing.GradientEvaluation that is stored in self.g_evaluation
        Also, computes the gradient and gradient_norm and saves them to Json .
        """
        assert self.target_LOW.VN == target_1.VN, "Attention! Preprocessor does not match targets. "
        assign_interpolation_distance_to_mesh_vertices(self.mesh, weight=0.5,
                                                       target_LOW=self.target_LOW, target_HIGH=self.target_HIGH)
        g_evaluation = GradientEvaluation(self.mesh, self.DATA_PATH)
        g_evaluation.compute_gradient()
        g_evaluation.compute_gradient_norm()

        if save_output:
            # save results to json
            utils.save_to_json(g_evaluation.vertex_gradient_norm, self.OUTPUT_PATH, norm_filename)
            utils.save_to_json(utils.point_list_to_dict(g_evaluation.vertex_gradient), self.OUTPUT_PATH, g_filename)

        return g_evaluation

    def find_critical_points(self, g_evaluation, output_filenames):
        """ Computes and saves to json the critical points of the df on the mesh (minima, maxima, saddles)"""
        g_evaluation.find_critical_points()
        # save results to json
        utils.save_to_json(g_evaluation.minima, self.OUTPUT_PATH, output_filenames[0])
        utils.save_to_json(g_evaluation.maxima, self.OUTPUT_PATH, output_filenames[1])
        utils.save_to_json(g_evaluation.saddles, self.OUTPUT_PATH, output_filenames[2])

    ###########################
    # --- Region Split

    def region_split(self, cut_mesh=True, separate_neighborhoods=True, topological_sorting=True,
                     save_split_meshes=True):
        """
        Splits the mesh on the saddle points. This process can take a long time.
        It consists of four parts:
        1) Create cuts on the mesh so that they intersect the saddle points and follow the get_distance function
        iso-contour
        2) Separate mesh neighborhoods  from cuts
        3) Topological sorting of split meshes to determine their connectivity and sequence.
        4) Finally resulting meshes are saved to json.

        The intermediary outputs are saved to json, so if you don'weight want to be recomputing the entire thing every
        time, you can turn the respective processes to false.
        """

        print("")
        logging.info("--- Mesh region splitting")

        if cut_mesh:  # (1)
            self.mesh.update_default_vertex_attributes({'cut': 0})
            mesh_splitter = rs.MeshSplitter(self.mesh, self.target_LOW, self.target_HIGH, self.DATA_PATH)
            mesh_splitter.run()

            self.mesh = mesh_splitter.mesh
            logger.info('Completed Region splitting')
            logger.info("Region split cut indices: " + str(mesh_splitter.cut_indices))
            # save results to json
            self.mesh.to_obj(os.path.join(self.OUTPUT_PATH, 'mesh_with_cuts.obj'))
            self.mesh.to_json(os.path.join(self.OUTPUT_PATH, 'mesh_with_cuts.json'))
            logger.info("Saving to Obj and Json: " + os.path.join(self.OUTPUT_PATH, 'mesh_with_cuts.json'))

        if separate_neighborhoods:  # (2)
            print("")
            logger.info("--- Separating mesh disconnected components")
            self.mesh = Mesh.from_json(os.path.join(self.OUTPUT_PATH, 'mesh_with_cuts.json'))
            region_split_cut_indices = get_existing_cut_indices(self.mesh)

            # save results to json
            utils.save_to_json(get_vertices_that_belong_to_cuts(self.mesh, region_split_cut_indices),
                               self.OUTPUT_PATH, "vertices_on_cuts.json")

            self.split_meshes = rs.separate_disconnected_components(self.mesh, attr='cut',
                                                                    values=region_split_cut_indices,
                                                                    OUTPUT_PATH=self.OUTPUT_PATH)
            logger.info('Created %d split meshes.' % len(self.split_meshes))

        if topological_sorting:  # (3)
            print("")
            logger.info("--- Topological sort of meshes directed graph to determine print order")
            graph = topo_sort.MeshDirectedGraph(self.split_meshes, self.DATA_PATH)
            all_orders = graph.get_all_topological_orders()
            selected_order = all_orders[0]
            logger.info('selected_order : ' + str(selected_order))  # TODO: improve the way an order is selected
            self.cleanup_mesh_attributes_based_on_selected_order(selected_order, graph)

            # reorder split_meshes based on selected order
            self.split_meshes = [self.split_meshes[i] for i in selected_order]

        # --- save split meshes
        if save_split_meshes:  # (4)
            print("")
            logger.info("--- Saving resulting split meshes")
            for i, m in enumerate(self.split_meshes):
                m.to_obj(os.path.join(self.OUTPUT_PATH, 'split_mesh_' + str(i) + '.obj'))
                m.to_json(os.path.join(self.OUTPUT_PATH, 'split_mesh_' + str(i) + '.json'))
            logger.info('Saving to Obj and Json: ' + os.path.join(self.OUTPUT_PATH, 'split_mesh_%.obj'))
            logger.info("Saved %d split_meshes" % len(self.split_meshes))
            print('')

    def cleanup_mesh_attributes_based_on_selected_order(self, selected_order, graph):
        """
        Based on the selected order of split meshes, it rearranges their attributes, so that they can then be used
        with an interpolation slicer that requires data['boundary'] to be filled for every vertex.
        The vertices that originated from cuts have data['cut']=cut_index. This is replaced
        by data['boundary'] = 1 or 2 depending on connectivity of mesh.

        Parameters
        ----------
        selected_order: list, int
            The indices of ordered split meshes.
        graph: :class: 'networkx.Graph'
        """
        for index in selected_order:
            mesh = self.split_meshes[index]
            for child_node in graph.adj_list[index]:
                child_mesh = self.split_meshes[child_node]
                edge = graph.G.edges[index, child_node]
                common_cuts = edge['cut']
                for cut_id in common_cuts:
                    replace_mesh_vertex_attribute(mesh, 'cut', cut_id, 'boundary', 2)
                    replace_mesh_vertex_attribute(child_mesh, 'cut', cut_id, 'boundary', 1)

            # save results to json
            pts_boundary_LOW = utils.get_mesh_vertex_coords_with_attribute(mesh, 'boundary', 1)
            pts_boundary_HIGH = utils.get_mesh_vertex_coords_with_attribute(mesh, 'boundary', 2)
            utils.save_to_json(utils.point_list_to_dict(pts_boundary_LOW), self.OUTPUT_PATH,
                               'pts_boundary_LOW_%d.json' % index)
            utils.save_to_json(utils.point_list_to_dict(pts_boundary_HIGH), self.OUTPUT_PATH,
                               'pts_boundary_HIGH_%d.json' % index)


if __name__ == "__main__":
    pass



import itertools
from compas.geometry import Point
from compas_slicer.geometry import Layer
from compas_slicer.geometry import Path
import progressbar
import logging
import compas_slicer.utilities as utils
from compas.plugins import PluginNotInstalledError

packages = utils.TerminalCommand('conda list').get_split_output_strings()

if 'compas-cgal' in packages or 'compas_cgal' in packages:
    from compas_cgal.slicer import slice_mesh

logger = logging.getLogger('logger')

__all__ = ['create_planar_paths_cgal']


def create_planar_paths_cgal(mesh, planes):
    """Creates planar contours using CGAL
    Considers all resulting paths as CLOSED paths.
    This is a very fast method.

    Parameters
    ----------
    mesh: :class: 'compas.datastructures.Mesh'
        A compas mesh.
    planes: list, :class: 'compas.geometry.Plane'
    """
    if 'compas-cgal' not in packages and 'compas_cgal' not in packages:
        raise PluginNotInstalledError("--------ATTENTION! ----------- \
                        Compas_cgal library is missing! \
                        You can't use this planar slicing method without it. \
                        Check the README instructions for how to install it, \
                        or use another planar slicing method.")

    # prepare mesh for slicing
    M = mesh.to_vertices_and_faces()

    # slicing operation
    contours = slice_mesh(M, planes)
    cgal_layers = get_grouped_list(contours, key_function=key_function)

    layers = []
    with progressbar.ProgressBar(max_value=len(planes)) as bar:
        for i, layer in enumerate(cgal_layers):
            paths_per_layer = []
            for path in layer:
                points_per_contour = []
                for point in path:
                    pt = Point(point[0], point[1], point[2])
                    points_per_contour.append(pt)
                # generate contours
                # TODO: add a check for is_closed
                path = Path(points=points_per_contour, is_closed=True)
                paths_per_layer.append(path)

            # generate layers
            layer = Layer(paths_per_layer)
            layers.append(layer)

            # advance progressbar
            bar.update(i)

    return layers


def get_grouped_list(item_list, key_function):
    """ Groups layers horizontally. """
    # first sort, because grouping only groups consecutively matching items
    sorted_list = sorted(item_list, key=key_function)
    # group items, using the provided key function
    grouped_iter = itertools.groupby(sorted_list, key_function)
    # return reformatted list
    return [list(group) for _key, group in grouped_iter]


def key_function(item):
    return item[0][2]


if __name__ == "__main__":
    pass



import logging
import itertools

from compas.geometry import Point, distance_point_point

logger = logging.getLogger('logger')

__all__ = ['reorder_vertical_layers']


def reorder_vertical_layers(slicer, align_with):
    """Re-orders the vertical layers in a specific way

    Parameters
    ----------
    slicer: :class:`compas_slicer.slicers.BaseSlicer`
        An instance of one of the compas_slicer.slicers classes.
    align_with: str or :class:`compas.geometry.Point`
        x_axis       = reorders the vertical layers starting from the positive x-axis
        y_axis       = reorders the vertical layers starting from the positive y-axis
        Point(x,y,z) = reorders the vertical layers starting from a given Point
    """

    if align_with == "x_axis":
        align_pt = Point(2 ** 32, 0, 0)
    elif align_with == "y_axis":
        align_pt = Point(0, 2 ** 32, 0)
    elif isinstance(align_with, Point):
        align_pt = align_with
    else:
        raise NameError("Unknown align_with : " + str(align_with))

    logger.info("Re-ordering vertical layers to start with the vertical layer closest to: %s" % align_with)

    for layer in slicer.layers:
        assert layer.min_max_z_height[0] is not None and layer.min_max_z_height[1] is not None, \
            "To use the 'reorder_vertical_layers function you need first to calculate the layers' z_bounds. To do " \
            "that use the function 'Layer.calculate_z_bounds()'"

    # group vertical layers based on the min_max_z_height
    grouped_iter = itertools.groupby(slicer.layers, lambda x: x.min_max_z_height)
    grouped_layer_list = [list(group) for _key, group in grouped_iter]

    reordered_layers = []

    for grouped_layers in grouped_layer_list:
        distances = []
        for vert_layer in grouped_layers:
            # recreate head_centroid_pt as compas.Point
            head_centroid_pt = Point(vert_layer.head_centroid[0], vert_layer.head_centroid[1], vert_layer.head_centroid[2])
            # measure distance
            distances.append(distance_point_point(head_centroid_pt, align_pt))

        # sort lists based on closest distance to align pt
        grouped_new = [x for _, x in sorted(zip(distances, grouped_layers))]
        reordered_layers.append(grouped_new)

    # flatten list
    slicer.layers = [item for sublist in reordered_layers for item in sublist]


if __name__ == "__main__":
    pass



import compas_slicer

__all__ = ['get_param']


def get_param(params, key, defaults_type):
    """
    Function useful for accessing the params dictionary of curved slicing.
    If the key is in the params dict, it returns its value,
    otherwise it returns the default_value.

    Parameters
    ----------
    params: dict
    key: str
    defaults_type: str specifying which defaults the dictionary of parameters draws for. 'curved_slicing' / 'gcode'

    Returns
    ----------
    params[key] if key in params, otherwise default_value
    """
    if key in params:
        return params[key]
    else:
        if defaults_type == 'interpolation_slicing':
            return compas_slicer.parameters.interpolation_slicing_default_param(key)
        elif defaults_type == 'gcode':
            return compas_slicer.parameters.gcode_default_param(key)
        elif defaults_type == 'layers':
            return compas_slicer.parameters.layers_default_param(key)
        elif defaults_type == 'print_organization':
            return compas_slicer.parameters.gcode_default_param(key)
        else:
            raise ValueError('The specified parameter type : ' + str(defaults_type) + ' does not exist.')



from compas.geometry import barycentric_coordinates
import logging
import progressbar
from compas_slicer.utilities.utils import pull_pts_to_mesh_faces

logger = logging.getLogger('logger')


__all__ = ['transfer_mesh_attributes_to_printpoints']


######################
# PrintPoints Attributes
######################

def transfer_mesh_attributes_to_printpoints(mesh, printpoints_dict):
    """
    Transfers face and vertex attributes from the mesh to the printpoints.
    Each printpoint is projected to the closest mesh face. It takes directly all the face attributes.
    It takes the averaged vertex attributes of the face vertices using barycentric coordinates.

    Face attributes can be anything.
    Vertex attributes can only be entities that support multiplication with a scalar. They have been tested to work
    with scalars and np.arrays.

    The reserved attribute names (see 'is_reserved_attribute(attr)') are not passed on to the printpoints.
    """
    logger.info('Transferring mesh attributes to the printpoints.')

    all_pts = []
    for layer_key in printpoints_dict:
        for path_key in printpoints_dict[layer_key]:
            for ppt in printpoints_dict[layer_key][path_key]:
                all_pts.append(ppt.pt)

    closest_fks, projected_pts = pull_pts_to_mesh_faces(mesh, all_pts)

    i = 0
    with progressbar.ProgressBar(max_value=len(all_pts)) as bar:
        for layer_key in printpoints_dict:
            for path_key in printpoints_dict[layer_key]:
                for ppt in printpoints_dict[layer_key][path_key]:
                    fkey = closest_fks[i]
                    proj_pt = projected_pts[i]
                    ppt.attributes = transfer_mesh_attributes_to_point(mesh, fkey, proj_pt)
                    i += 1
                    bar.update(i)


def is_reserved_attribute(attr):
    """ Returns True if the attribute name is a reserved, false otherwise. """
    taken_attributes = ['x', 'y', 'z', 'uv',
                        'scalar_field']
    return attr in taken_attributes


def transfer_mesh_attributes_to_point(mesh, fkey, proj_pt):
    """
    It projects the point on the closest face of the mesh. Then if finds
    all the vertex and face attributes of the face and its attributes and transfers them to the point.
    The vertex attributes are transferred using barycentric coordinates.

    Parameters
    ----------
    mesh: compas.datastructures.Mesh
    fkey: face key
    proj_pt: list [x,y,z], point projected on the plane of the face fkey

    Returns
    -------
    dict that contains all the attributes that correspond to the printpoints position
    """

    vs = mesh.face_vertices(fkey)
    bar_coords = barycentric_coordinates(proj_pt, triangle=(mesh.vertex_coordinates(vs[0]),
                                                            mesh.vertex_coordinates(vs[1]),
                                                            mesh.vertex_coordinates(vs[2])))

    # get face attributes
    face_attrs = mesh.face_attributes(fkey)
    keys_to_remove = [attr for attr in face_attrs if is_reserved_attribute(attr)]
    for key in keys_to_remove:
        del face_attrs[key]  # remove from face_attrs dictionary

    # get vertex attributes using barycentric coordinates
    vs = mesh.face_vertices(fkey)
    vertex_attrs = {}
    checked_attrs = []
    for attr in mesh.vertex_attributes(vs[0]):
        if not is_reserved_attribute(attr):
            if not (attr in checked_attrs):
                check_that_attribute_can_be_multiplied(attr, mesh.vertex_attributes(vs[0])[attr])
                checked_attrs.append(attr)
            vertex_attrs[attr] = 0
            vertex_attrs[attr] += bar_coords[0] * mesh.vertex_attributes(vs[0])[attr]
            vertex_attrs[attr] += bar_coords[1] * mesh.vertex_attributes(vs[1])[attr]
            vertex_attrs[attr] += bar_coords[2] * mesh.vertex_attributes(vs[2])[attr]

    vertex_attrs.update(face_attrs)  # merge two dictionaries
    return vertex_attrs


def check_that_attribute_can_be_multiplied(attr_name, value):
    try:
        value * 1.0
        return True
    except TypeError:
        raise ValueError('Attention! The following vertex attribute cannot be multiplied with a scalar. %s : %s '
                         % (attr_name, str(type(value))))



import compas_slicer
import logging

logger = logging.getLogger('logger')

__all__ = ['set_extruder_toggle',
           'override_extruder_toggle',
           'check_assigned_extruder_toggle']


def set_extruder_toggle(print_organizer, slicer):
    """Sets the extruder_toggle value for the printpoints.

    Parameters
    ----------
    print_organizer: :class:`compas_slicer.print_organization.BasePrintOrganizer`
    slicer: :class:`compas.slicers.BaseSlicer`
    """
    logger.info("Setting extruder toggle")

    pp_dict = print_organizer.printpoints_dict

    for i, layer in enumerate(slicer.layers):
        layer_key = 'layer_%d' % i
        is_vertical_layer = isinstance(layer, compas_slicer.geometry.VerticalLayer)
        is_brim_layer = layer.is_brim

        for j, path in enumerate(layer.paths):
            path_key = 'path_%d' % j
            is_closed_path = path.is_closed

            # --- decide if the path should be interrupted at the end
            interrupt_path = False

            if not is_closed_path:
                interrupt_path = True
                # open paths should always be interrupted

            if not is_vertical_layer and len(layer.paths) > 1:
                interrupt_path = True
                # horizontal layers with multiple paths should be interrupted so that the extruder
                # can travel from one path to the other, exception is added for the brim layers
                if is_brim_layer and (j + 1) % layer.number_of_brim_offsets != 0:
                    interrupt_path = False

            if is_vertical_layer and j == len(layer.paths) - 1:
                interrupt_path = True
                # the last path of a vertical layer should be interrupted

            # --- create extruder toggles
            path_printpoints = pp_dict[layer_key][path_key]
            for k, printpoint in enumerate(path_printpoints):

                if interrupt_path:
                    if k == len(path_printpoints) - 1:
                        printpoint.extruder_toggle = False
                    else:
                        printpoint.extruder_toggle = True
                else:
                    printpoint.extruder_toggle = True

        # set extruder toggle of last print point to false
        last_layer_key = 'layer_%d' % (len(pp_dict) - 1)
        last_path_key = 'path_%d' % (len(pp_dict[last_layer_key]) - 1)
        pp_dict[last_layer_key][last_path_key][-1].extruder_toggle = False


def override_extruder_toggle(print_organizer, override_value):
    """Overrides the extruder_toggle value for the printpoints with a user-defined value.

    Parameters
    ----------
    print_organizer: :class:`compas.print_organization.BasePrintOrganizer`
    override_value: bool
        Value to override the extruder_toggle values with.

    """
    assert isinstance(override_value, bool), "Override value must be of type bool"
    for printpoint in print_organizer.printpoints_iterator():
        printpoint.extruder_toggle = override_value


def check_assigned_extruder_toggle(print_organizer):
    """ Checks that all the printpoints have an assigned extruder toggle. """
    all_toggles_assigned = True
    for printpoint in print_organizer.printpoints_iterator():
        if printpoint.extruder_toggle is None:
            all_toggles_assigned = False
    return all_toggles_assigned


if __name__ == "__main__":
    pass



import numpy as np
import logging
import compas_slicer.utilities as utils
from compas_slicer.pre_processing.preprocessing_utils import get_face_gradient_from_scalar_field
from compas_slicer.pre_processing.preprocessing_utils import get_vertex_gradient_from_face_gradient

logger = logging.getLogger('logger')

__all__ = ['GradientEvaluation']


class GradientEvaluation(object):
    """
    Evaluation of the gradient of the scalar function of the mesh.
    The scalar function should be stored as a vertex attribute on every vertex, with key='scalar_field'

    Attributes
    ----------
    mesh: :class: 'compas.datastructures.Mesh'
    DATA_PATH: str, path to the data folder

    """
    def __init__(self, mesh, DATA_PATH):
        for v_key, data in mesh.vertices(data=True):
            assert 'scalar_field' in data, "Vertex %d does not have the attribute 'scalar_field'"

        print('')
        logger.info('Gradient evaluation')
        self.mesh = mesh
        self.DATA_PATH = DATA_PATH
        self.OUTPUT_PATH = utils.get_output_directory(DATA_PATH)

        self.minima, self.maxima, self.saddles = [], [], []

        self.face_gradient = []  # np.array (#F x 3) one gradient vector per face.
        self.vertex_gradient = []  # np.array (#V x 3) one gradient vector per vertex.
        self.face_gradient_norm = []  # list (#F x 1)
        self.vertex_gradient_norm = []  # list (#V x 1)

    def compute_gradient(self):
        """ Computes the gradient on the faces and the vertices. """
        u_v = [self.mesh.vertex[vkey]['scalar_field'] for vkey in self.mesh.vertices()]
        self.face_gradient = get_face_gradient_from_scalar_field(self.mesh, u_v)
        self.vertex_gradient = get_vertex_gradient_from_face_gradient(self.mesh, self.face_gradient)

    def compute_gradient_norm(self):
        """ Computes the norm of the gradient. """
        logger.info('Computing norm of gradient')
        f_g = np.array([self.face_gradient[i] for i, fkey in enumerate(self.mesh.faces())])
        v_g = np.array([self.vertex_gradient[i] for i, vkey in enumerate(self.mesh.vertices())])
        self.face_gradient_norm = list(np.linalg.norm(f_g, axis=1))
        self.vertex_gradient_norm = list(np.linalg.norm(v_g, axis=1))

    def find_critical_points(self):
        """ Finds minima, maxima and saddle points of the scalar function on the mesh. """
        for vkey, data in self.mesh.vertices(data=True):
            current_v = data['scalar_field']
            neighbors = self.mesh.vertex_neighbors(vkey, ordered=True)
            values = []
            if len(neighbors) > 0:
                neighbors.append(neighbors[0])
                for n in neighbors:
                    v = self.mesh.vertex_attributes(n)['scalar_field']
                    if abs(v - current_v) > 0.0:
                        values.append(current_v - v)
                sgc = count_sign_changes(values)

                if sgc == 0:  # extreme point
                    if current_v > self.mesh.vertex_attributes(neighbors[0])['scalar_field']:
                        self.maxima.append(vkey)
                    else:
                        self.minima.append(vkey)
                if sgc == 2:  # regular point
                    pass
                if sgc > 2 and sgc % 2 == 0:
                    self.saddles.append(vkey)


#####################################
# --- Helpers


def count_sign_changes(values):
    """ Returns the number of sign changes in a list of values. """
    count = 0
    prev_v = 0
    for i, v in enumerate(values):
        if i == 0:
            prev_v = v
        else:
            if prev_v * v < 0:
                count += 1
            prev_v = v
    return count


if __name__ == "__main__":
    pass



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from .safety_printpoints import *  # noqa: F401 E402 F403
from .blend_radius import *  # noqa: F401 E402 F403
from .linear_velocity import *  # noqa: F401 E402 F403
from .extruder_toggle import *  # noqa: F401 E402 F403
from .wait_time import *  # noqa: F401 E402 F403
from .gcode import *  # noqa: F401 E402 F403
from .data_smoothing import *  # noqa: F401 E402 F403


__all__ = [name for name in dir() if not name.startswith('_')]



from compas.geometry import Vector
from compas_slicer.print_organization.print_organization_utilities.extruder_toggle import check_assigned_extruder_toggle
from compas_slicer.utilities import find_next_printpoint
import copy
import logging

logger = logging.getLogger('logger')

__all__ = ['add_safety_printpoints']


def add_safety_printpoints(print_organizer, z_hop=10.0):
    """Generates a safety print point at the interruptions of the print paths.

    Parameters
    ----------
    print_organizer: :class:`compas_slicer.print_organization.BasePrintOrganizer`
        An instance of the BasePrintOrganizer class.
    z_hop: float
        Vertical distance (in millimeters) of the safety point above the PrintPoint.
    """
    assert check_assigned_extruder_toggle(print_organizer), \
        'You need to set the extruder toggles first, before you can create safety points'
    logger.info("Generating safety print points with height " + str(z_hop) + " mm")

    pp_dict = print_organizer.printpoints_dict
    pp_copy_dict = {}  # should not be altering the dict that we are iterating through > copy

    for i, layer_key in enumerate(pp_dict):
        pp_copy_dict[layer_key] = {}

        for j, path_key in enumerate(pp_dict[layer_key]):
            pp_copy_dict[layer_key][path_key] = []

            for k, printpoint in enumerate(pp_dict[layer_key][path_key]):
                #  add regular printing points
                pp_copy_dict[layer_key][path_key].append(printpoint)

                # add safety printpoints if there is an interruption
                if printpoint.extruder_toggle is False:

                    # safety ppt after current printpoint
                    pp_copy_dict[layer_key][path_key].append(create_safety_printpoint(printpoint, z_hop, False))

                    #  safety ppt before next printpoint (if there exists one)
                    next_ppt = find_next_printpoint(pp_dict, i, j, k)
                    if next_ppt:
                        if next_ppt.extruder_toggle is True:  # if it is a printing ppt
                            pp_copy_dict[layer_key][path_key].append(create_safety_printpoint(next_ppt, z_hop, False))

    #  finally, insert a safety print point at the beginning of the entire print
    safety_printpoint = create_safety_printpoint(pp_dict['layer_0']['path_0'][0], z_hop, False)
    pp_copy_dict['layer_0']['path_0'].insert(0, safety_printpoint)

    #  the safety printpoint has already been added at the end since the last printpoint extruder_toggle_type is False
    print_organizer.printpoints_dict = pp_copy_dict


def create_safety_printpoint(printpoint, z_hop, extruder_toggle):
    """

    Parameters
    ----------
    printpoint: :class: 'compas_slicer.geometry.PrintPoint'
    z_hop: float
    extruder_toggle: bool

    Returns
    ----------
    :class: 'compas_slicer.geometry.PrintPoint'
    """

    pt0 = printpoint.pt
    safety_printpoint = copy.deepcopy(printpoint)
    safety_printpoint.pt = pt0 + Vector(0, 0, z_hop)
    safety_printpoint.extruder_toggle = extruder_toggle
    return safety_printpoint


if __name__ == "__main__":
    pass



from compas.geometry import Vector, normalize_vector
from compas_slicer.print_organization import BasePrintOrganizer
import compas_slicer.utilities as utils
from compas_slicer.geometry import PrintPoint
import progressbar
import logging
from compas_slicer.pre_processing import GradientEvaluation
from compas_slicer.utilities.attributes_transfer import transfer_mesh_attributes_to_printpoints
from compas_slicer.parameters import get_param
import compas_slicer

logger = logging.getLogger('logger')

__all__ = ['ScalarFieldPrintOrganizer']


class ScalarFieldPrintOrganizer(BasePrintOrganizer):
    """
    Organizing the printing process for the realization of planar contours.

    Attributes
    ----------
    slicer: :class:`compas_slicer.slicers.PlanarSlicer`
        An instance of the compas_slicer.slicers.PlanarSlicer.
    """

    def __init__(self, slicer, parameters, DATA_PATH):
        assert isinstance(slicer, compas_slicer.slicers.ScalarFieldSlicer), 'Please provide a ScalarFieldSlicer'
        BasePrintOrganizer.__init__(self, slicer)
        self.DATA_PATH = DATA_PATH
        self.OUTPUT_PATH = utils.get_output_directory(DATA_PATH)
        self.parameters = parameters

        self.vertical_layers = slicer.vertical_layers
        self.horizontal_layers = slicer.horizontal_layers
        assert len(self.vertical_layers) + len(self.horizontal_layers) == len(slicer.layers)

        if len(self.horizontal_layers) > 0:
            assert len(self.horizontal_layers) == 1, "Only one brim horizontal layer is currently supported."
            assert self.horizontal_layers[0].is_brim, "Only one brim horizontal layer is currently supported."
            logger.info('Slicer has one horizontal brim layer.')

        self.g_evaluation = self.add_gradient_to_vertices()

    def __repr__(self):
        return "<ScalarFieldPrintOrganizer with %i layers>" % len(self.slicer.layers)

    def create_printpoints(self):
        """ Create the print points of the fabrication process """
        count = 0
        logger.info('Creating print points ...')
        with progressbar.ProgressBar(max_value=self.slicer.number_of_points) as bar:

            for i, layer in enumerate(self.slicer.layers):
                self.printpoints_dict['layer_%d' % i] = {}

                for j, path in enumerate(layer.paths):
                    self.printpoints_dict['layer_%d' % i]['path_%d' % j] = []

                    for k, point in enumerate(path.points):
                        normal = utils.get_normal_of_path_on_xy_plane(k, point, path, self.slicer.mesh)

                        h = get_param(self.parameters, 'avg_layer_height', defaults_type='layers')
                        printpoint = PrintPoint(pt=point, layer_height=h, mesh_normal=normal)

                        self.printpoints_dict['layer_%d' % i]['path_%d' % j].append(printpoint)
                        bar.update(count)
                        count += 1

        # transfer gradient information to printpoints
        transfer_mesh_attributes_to_printpoints(self.slicer.mesh, self.printpoints_dict)

        # add non-planar print data to printpoints
        for i, layer in enumerate(self.slicer.layers):
            layer_key = 'layer_%d' % i
            for j, path in enumerate(layer.paths):
                path_key = 'path_%d' % j
                for pp in self.printpoints_dict[layer_key][path_key]:
                    grad_norm = pp.attributes['gradient_norm']
                    grad = pp.attributes['gradient']
                    pp.distance_to_support = grad_norm
                    pp.layer_height = grad_norm
                    pp.up_vector = Vector(*normalize_vector(grad))
                    pp.frame = pp.get_frame()

    def add_gradient_to_vertices(self):
        g_evaluation = GradientEvaluation(self.slicer.mesh, self.DATA_PATH)
        g_evaluation.compute_gradient()
        g_evaluation.compute_gradient_norm()

        utils.save_to_json(g_evaluation.vertex_gradient_norm, self.OUTPUT_PATH, 'gradient_norm.json')
        utils.save_to_json(utils.point_list_to_dict(g_evaluation.vertex_gradient), self.OUTPUT_PATH, 'gradient.json')

        self.slicer.mesh.update_default_vertex_attributes({'gradient': 0.0})
        self.slicer.mesh.update_default_vertex_attributes({'gradient_norm': 0.0})
        for i, (v_key, data) in enumerate(self.slicer.mesh.vertices(data=True)):
            data['gradient'] = g_evaluation.vertex_gradient[i]
            data['gradient_norm'] = g_evaluation.vertex_gradient_norm[i]
        return g_evaluation


if __name__ == "__main__":
    pass



from compas_slicer.slicers import BaseSlicer
import logging
from compas_slicer.slicers.slice_utilities import UVContours
import numpy as np

import progressbar
from compas_slicer.geometry import VerticalLayersManager
from compas_slicer.parameters import get_param

logger = logging.getLogger('logger')

__all__ = ['UVSlicer']


class UVSlicer(BaseSlicer):
    """
    Generates the contours on the mesh that correspond to straight lines on the plane,
    using on a UV map (from 3D space to the plane) defined on the mesh vertices.

    Attributes
    ----------
    mesh: :class: 'compas.datastructures.Mesh'
        Input mesh, it must be a triangular mesh (i.e. no quads or n-gons allowed)
        Note that the topology of the mesh matters, irregular tesselation can lead to undesired results.
        We recommend to 1)re-topologize, 2) triangulate, and 3) weld your mesh in advance.
    vkey_to_uv: dict {vkey : tuple (u,v)}. U,V coordinates should be in the domain [0,1]. The U coordinate
    no_of_isocurves: int, how many levels to be generated
    """

    def __init__(self, mesh, vkey_to_uv, no_of_isocurves, parameters=None):
        logger.info('UVSlicer')
        BaseSlicer.__init__(self, mesh)

        self.vkey_to_uv = vkey_to_uv
        self.no_of_isocurves = no_of_isocurves
        self.parameters = parameters if parameters else {}

        u = [self.vkey_to_uv[vkey][0] for vkey in mesh.vertices()]
        v = [self.vkey_to_uv[vkey][1] for vkey in mesh.vertices()]
        u = np.array(u) * float(no_of_isocurves + 1)
        vkey_to_i = self.mesh.key_index()

        mesh.update_default_vertex_attributes({'uv': 0})
        for vkey in mesh.vertices():
            mesh.vertex_attribute(vkey, 'uv', (u[vkey_to_i[vkey]], v[vkey_to_i[vkey]]))

    def generate_paths(self):
        """ Generates isocontours. """
        paths_type = 'flat'  # 'spiral' # 'zigzag'
        v_left, v_right = 0.0, 1.0 - 1e-5

        max_dist = get_param(self.parameters, key='vertical_layers_max_centroid_dist', defaults_type='layers')
        vertical_layers_manager = VerticalLayersManager(max_dist)

        # create paths + layers
        with progressbar.ProgressBar(max_value=self.no_of_isocurves) as bar:
            for i in range(0, self.no_of_isocurves + 1):
                if i == 0:
                    i += 0.05  # contours are a bit tricky in the edges
                if paths_type == 'spiral':
                    u1, u2 = i, i + 1.0
                else:  # 'flat'
                    u1 = u2 = i

                p1 = (u1, v_left)
                p2 = (u2, v_right)

                contours = UVContours(self.mesh, p1, p2)
                contours.compute()
                contours.add_to_vertical_layers_manager(vertical_layers_manager)

                bar.update(i)  # advance progress bar

        self.layers = vertical_layers_manager.layers


if __name__ == "__main__":
    pass



import numpy as np
import logging
import scipy

logger = logging.getLogger('logger')

__all__ = ['get_vertex_gradient_from_face_gradient',
           'get_edge_gradient_from_vertex_gradient',
           'get_face_gradient_from_scalar_field',
           'normalize_gradient',
           'get_per_vertex_divergence',
           'get_scalar_field_from_gradient']


def get_vertex_gradient_from_face_gradient(mesh, face_gradient):
    """
    Finds vertex gradient given an already calculated per face gradient.

    Parameters
    ----------
    mesh: :class: 'compas.datastructures.Mesh'
    face_gradient: np.array with one vec3 per face of the mesh. (dimensions : #F x 3)

    Returns
    ----------
    np.array (dimensions : #V x 3) one gradient vector per vertex.
    """
    logger.info('Computing per vertex gradient')
    vertex_gradient = []
    for v_key in mesh.vertices():
        faces_total_area = 0
        faces_total_grad = np.array([0.0, 0.0, 0.0])
        for f_key in mesh.vertex_faces(v_key):
            face_area = mesh.face_area(f_key)
            faces_total_area += face_area
            faces_total_grad += face_area * face_gradient[f_key, :]
        v_grad = faces_total_grad / faces_total_area
        vertex_gradient.append(v_grad)
    return np.array(vertex_gradient)


def get_edge_gradient_from_vertex_gradient(mesh, vertex_gradient):
    """
    Finds edge gradient given an already calculated per vertex gradient.

    Parameters
    ----------
    mesh: :class: 'compas.datastructures.Mesh'
    vertex_gradient: np.array with one vec3 per vertex of the mesh. (dimensions : #V x 3)

    Returns
    ----------
    np.array (dimensions : #E x 3) one gradient vector per edge.
    """
    edge_gradient = []
    for u, v in mesh.edges():
        thisEdgeGradient = vertex_gradient[u] + vertex_gradient[v]
        edge_gradient.append(thisEdgeGradient)
    return np.array(edge_gradient)


def get_face_gradient_from_scalar_field(mesh, u, use_igl=True):
    """
    Finds face gradient from scalar field u.
    Scalar field u is given per vertex.

    Parameters
    ----------
    mesh: :class: 'compas.datastructures.Mesh'
    u: list, float. (dimensions : #VN x 1)

    Returns
    ----------
    np.array (dimensions : #F x 3) one gradient vector per face.
    """
    logger.info('Computing per face gradient')
    if use_igl:
        import igl
        v, f = mesh.to_vertices_and_faces()
        G = igl.grad(np.array(v), np.array(f))
        X = G * u
        nf = len(list(mesh.faces()))
        X = np.array([[X[i], X[i + nf], X[i + 2 * nf]] for i in range(nf)])
        return X
    else:
        grad = []
        for fkey in mesh.faces():
            A = mesh.face_area(fkey)
            N = mesh.face_normal(fkey)
            edge_0, edge_1, edge_2 = get_face_edge_vectors(mesh, fkey)
            v0, v1, v2 = mesh.face_vertices(fkey)
            u0 = u[v0]
            u1 = u[v1]
            u2 = u[v2]
            vc0 = np.array(mesh.vertex_coordinates(v0))
            vc1 = np.array(mesh.vertex_coordinates(v1))
            vc2 = np.array(mesh.vertex_coordinates(v2))
            # grad_u = -1 * ((u1-u0) * np.cross(vc0-vc2, N) + (u2-u0) * np.cross(vc1-vc0, N)) / (2 * A)
            grad_u = ((u1-u0) * np.cross(vc0-vc2, N) + (u2-u0) * np.cross(vc1-vc0, N)) / (2 * A)
            # grad_u = (np.cross(N, edge_0) * u2 +
            #           np.cross(N, edge_1) * u0 +
            #           np.cross(N, edge_2) * u1) / (2 * A)
            grad.append(grad_u)
        return np.array(grad)


def get_face_edge_vectors(mesh, fkey):
    """ Returns the edge vectors of the face with fkey. """
    e0, e1, e2 = mesh.face_halfedges(fkey)
    edge_0 = np.array(mesh.vertex_coordinates(e0[0])) - np.array(mesh.vertex_coordinates(e0[1]))
    edge_1 = np.array(mesh.vertex_coordinates(e1[0])) - np.array(mesh.vertex_coordinates(e1[1]))
    edge_2 = np.array(mesh.vertex_coordinates(e2[0])) - np.array(mesh.vertex_coordinates(e2[1]))
    return edge_0, edge_1, edge_2


def get_per_vertex_divergence(mesh, X, cotans):
    """
    Computes the divergence of the gradient X for the mesh, using cotangent weights.

    Parameters
    ----------
    mesh: :class: 'compas.datastructures.Mesh'
    X: np.array, (dimensions: #F x 3), per face gradient
    cotans:  np.array, (dimensions: #F x 3), 1/2*cotangents corresponding angles

    Returns
    ----------
    np.array (dimensions : #V x 1) one float (divergence value) per vertex.
    """
    cotans = cotans.reshape(-1, 3)
    div_X = np.zeros(len(list(mesh.vertices())))
    for fi, fkey in enumerate(mesh.faces()):
        x_fi = X[fi]
        edges = np.array(get_face_edge_vectors(mesh, fkey))
        for i in range(3):
            j = (i + 1) % 3
            k = (i + 2) % 3
            div_X[mesh.face_vertices(fkey)[i]] += cotans[fi, k] * np.dot(x_fi, edges[i]) / 2.0
            div_X[mesh.face_vertices(fkey)[i]] += cotans[fi, j] * np.dot(x_fi, -edges[k]) / 2.0
    return div_X


def normalize_gradient(X):
    """ Returns normalized gradient X. """
    norm = np.linalg.norm(X, axis=1)[..., np.newaxis]
    return X / norm  # normalize


def get_scalar_field_from_gradient(mesh, X, C, cotans):
    """
    Find scalar field u that best explains gradient X.
    Laplacian(u) = Divergence(X).
    This defines a scalar field up to translation, then we subtract the min to make sure it starts from 0.

    Parameters
    ----------
    mesh: :class: 'compas.datastructures.Mesh'
    X: np.array, (dimensions: #F x 3), per face gradient
    C: 'scipy.sparse.csr_matrix',
        sparse matrix (dimensions: #V x #V), cotmatrix, each row i corresponding to v(i, :)
    cotans: np.array, (dimensions: #F x 3), 1/2*cotangents corresponding angles

    Returns
    ----------
    np.array (dimensions : #V x 1) one scalar value per vertex.
    """
    div_X = get_per_vertex_divergence(mesh, X, cotans)
    u = scipy.sparse.linalg.spsolve(C, div_X)
    logger.info('Solved Δ(u) = div(X). Linear system error |Δ(u) - div(X)| = ' + str(np.linalg.norm(C * u - div_X)))
    u = u - np.amin(u)  # make start value equal 0
    u = 2*u
    return u


if __name__ == "__main__":
    pass



"""
********************************************************************************
pre_processing
********************************************************************************

Positioning
=======================

.. autosummary::
    :toctree: generated/
    :nosignatures:

    move_mesh_to_point
    get_mid_pt_base

"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from .preprocessing_utils import *  # noqa: F401 E402 F403
from .interpolation_slicing_preprocessor import *  # noqa: F401 E402 F403
from .gradient_evaluation import *  # noqa: F401 E402 F403

#  Positioning
from .positioning import *  # noqa: F401 E402 F403



from compas.geometry import Point, Frame, Vector, cross_vectors
import compas_slicer.utilities.utils as utils
import compas

__all__ = ['PrintPoint']


class PrintPoint(object):
    """
    A PrintPoint consists of a compas.geometry.Point,
    and additional attributes related to the printing process.

    Attributes
    ----------
    pt: :class:`compas.geometry.Point`
        A compas Point consisting out of x, y, z coordinates.
    layer_height: float
        The distance between the point on this layer and the previous layer.
        For planar slicing this is the vertical distance, for curved slicing this is absolute distance.
    mesh_normal: :class:`compas.geometry.Vector`
        Normal of the mesh at this PrintPoint.
        For planar_slicing it is the projection of the normal on the XY plane.
        # TODO: should we change that?
    up_vector: :class:`compas.geometry.Vector`
        Vector in up direction. For planar slicing this corresponds to the z axis, for curved slicing it varies.
    frame: :class:`compas.geometry.Frame`
        Frame with x-axis pointing up, y-axis pointing towards the mesh normal.
    extruder_toggle: bool
        True if extruder should be on (when printing), False if it should be off (when travelling).
    velocity: float
        Velocity to use for printing (print speed), in mm/s.
    wait_time: float
        Time in seconds to wait at this PrintPoint.
    """

    def __init__(self, pt, layer_height, mesh_normal):
        assert isinstance(pt, compas.geometry.Point)
        assert isinstance(mesh_normal, compas.geometry.Vector)
        assert layer_height

        #  --- basic printpoint
        self.pt = pt
        self.layer_height = layer_height

        self.mesh_normal = mesh_normal  # compas.geometry.Vector
        self.up_vector = Vector(0, 0, 1)  # default value that can be updated
        self.frame = self.get_frame()  # compas.geometry.Frame

        #  --- attributes transferred from the mesh (vertex / face attributes)
        self.attributes = {}  # dict. To fill this in,

        #  --- print_organization related attributes
        self.extruder_toggle = None  # bool
        self.velocity = None  # float (mm/s)
        self.wait_time = None  # float (sec)
        self.blend_radius = None  # float (mm)

        #  --- relation to support
        self.closest_support_pt = None  # <compas.geometry.Point>
        self.distance_to_support = None  # float

        self.is_feasible = True  # bool

    def __repr__(self):
        x, y, z = self.pt[0], self.pt[1], self.pt[2]
        return "<PrintPoint object at (%.2f, %.2f, %.2f)>" % (x, y, z)

    def get_frame(self):
        """ Returns a Frame with x-axis pointing up, y-axis pointing towards the mesh normal. """
        c = cross_vectors(self.up_vector, self.mesh_normal)
        return Frame(self.pt, c, self.mesh_normal)

    #################################
    #  --- To data , from data
    def to_data(self):
        """Returns a dictionary of structured data representing the data structure.
        TODO: The attributes of the printpoints are not saved in the dictionary because they can be non-Json
        serializable. Find a solution for this.

        Returns
        -------
        dict
            The PrintPoints' data.

        """
        point = {
            'point': [self.pt[0], self.pt[1], self.pt[2]],
            'layer_height': self.layer_height,

            'mesh_normal': self.mesh_normal.to_data(),
            'up_vector': self.up_vector.to_data(),
            'frame': self.frame.to_data(),

            'extruder_toggle': self.extruder_toggle,
            'velocity': self.velocity,
            'wait_time': self.wait_time,
            'blend_radius': self.blend_radius,

            'closest_support_pt': self.closest_support_pt.to_data() if self.closest_support_pt else None,
            'distance_to_support': self.distance_to_support,

            'is_feasible': self.is_feasible,

            'attributes': utils.get_jsonable_attributes(self.attributes)
        }
        return point

    @classmethod
    def from_data(cls, data):
        """Construct a PrintPoint from its data representation.

        Parameters
        ----------
        data: dict
            The data dictionary.

        Returns
        -------
        layer
            The constructed PrintPoint.

        """

        pp = cls(pt=Point.from_data(data['point']),
                 layer_height=data['layer_height'],
                 mesh_normal=Vector.from_data(data['mesh_normal']))

        pp.up_vector = Vector.from_data(data['up_vector'])
        pp.frame = Frame.from_data(data['frame'])

        pp.extruder_toggle = data['extruder_toggle']
        pp.velocity = data['velocity']
        pp.wait_time = data['wait_time']
        pp.blend_radius = data['blend_radius']

        pp.closest_support_pt = Point.from_data(data['closest_support_pt'])
        pp.distance_to_support = data['distance_to_support']

        pp.is_feasible = data['is_feasible']

        pp.attributes = data['attributes']
        return pp


if __name__ == "__main__":
    pass



import numpy as np
import math
from compas.datastructures import Mesh
import compas_slicer.utilities as utils
import logging
import networkx as nx
from compas_slicer.slicers.slice_utilities import create_graph_from_mesh_vkeys
from compas_slicer.pre_processing.preprocessing_utils.geodesics import get_igl_EXACT_geodesic_distances, \
    get_custom_HEAT_geodesic_distances

import statistics

logger = logging.getLogger('logger')

__all__ = ['CompoundTarget',
           'blend_union_list']


class CompoundTarget:
    """
    Represents a desired user-provided target. It acts as a key-frame that controls the print paths
    orientations. After the curved slicing , the print paths will be aligned to the compound target close to
    its area. The vertices that belong to the target are marked with their vertex attributes; they have
    data['v_attr'] = value.

    Attributes
    ----------
    mesh: :class:`compas.datastructures.Mesh`
    v_attr : str
        The key of the attribute dict to be checked.
    value: int
        The value of the attribute dict with key=v_attr. If in a vertex data[v_attr]==value then the vertex is part of
        this target.
    DATA_PATH: str
    has_blend_union: bool
    blend_radius : float
    geodesics_method: str
        'exact_igl'  exact igl geodesic distances
        'heat'   custom heat geodesic distances
    anisotropic_scaling: bool
        This is not yet implemented
    """

    def __init__(self, mesh, v_attr, value, DATA_PATH, has_blend_union=False, blend_radius=15.0,
                 geodesics_method='exact_igl', anisotropic_scaling=False):

        logger.info('Creating target with attribute : ' + v_attr + '=%d' % value)
        logger.info('has_blend_union : ' + str(has_blend_union) + ', blend_radius =  %.4f' % blend_radius)
        self.mesh = mesh
        self.v_attr = v_attr
        self.value = value
        self.DATA_PATH = DATA_PATH
        self.OUTPUT_PATH = utils.get_output_directory(DATA_PATH)

        self.has_blend_union = has_blend_union
        self.blend_radius = blend_radius

        self.geodesics_method = geodesics_method
        self.anisotropic_scaling = anisotropic_scaling  # Anisotropic scaling not yet implemented

        self.offset = 0
        self.VN = len(list(self.mesh.vertices()))

        # filled in by function 'self.find_targets_connected_components()'
        self.all_target_vkeys = []  # flattened list with all vi_starts
        self.clustered_vkeys = []  # nested list with all vi_starts
        self.number_of_boundaries = None  # int

        self.weight_max_per_cluster = []

        # geodesic distances
        # filled in by function 'self.update_distances_lists()'
        self._distances_lists = []  # nested list. Shape: number_of_boundaries x number_of_vertices
        self._distances_lists_flipped = []  # nested list. Shape: number_of_vertices x number_of_boundaries
        self._np_distances_lists_flipped = np.array([])  # numpy array of self._distances_lists_flipped
        self._max_dist = None  # maximum get_distance value from the target on any vertex of the mesh

        # compute
        self.find_targets_connected_components()
        self.compute_geodesic_distances()

    #  --- Neighborhoods clustering
    def find_targets_connected_components(self):
        """
        Clusters all the vertices that belong to the target into neighborhoods using a graph.
        Each target can have an arbitrary number of neighborhoods/clusters.
        Fills in the attributes: self.all_target_vkeys, self.clustered_vkeys, self.number_of_boundaries
        """
        self.all_target_vkeys = [vkey for vkey, data in self.mesh.vertices(data=True) if
                                 data[self.v_attr] == self.value]
        assert len(self.all_target_vkeys) > 0, "There are no vertices in the mesh with the attribute : " \
                                               + self.v_attr + ", value : %d" % self.value + " .Probably you made a " \
                                                                                             "mistake while creating the targets. "
        G = create_graph_from_mesh_vkeys(self.mesh, self.all_target_vkeys)
        assert len(list(G.nodes())) == len(self.all_target_vkeys)
        self.number_of_boundaries = len(list(nx.connected_components(G)))

        for i, cp in enumerate(nx.connected_components(G)):
            self.clustered_vkeys.append(list(cp))
        logger.info("Compound target with 'boundary'=%d. Number of connected_components : %d" % (
            self.value, len(list(nx.connected_components(G)))))

    #  --- Geodesic distances
    def compute_geodesic_distances(self):
        """
        Computes the geodesic distances from each of the target's neighborhoods  to all the mesh vertices.
        Fills in the distances attributes.
        """
        if self.geodesics_method == 'exact_igl':
            distances_lists = [get_igl_EXACT_geodesic_distances(self.mesh, vstarts) for vstarts in
                               self.clustered_vkeys]
        elif self.geodesics_method == 'heat':
            distances_lists = [get_custom_HEAT_geodesic_distances(self.mesh, vstarts, self.OUTPUT_PATH) for vstarts in
                               self.clustered_vkeys]
        else:
            raise ValueError('Unknown geodesics method : ' + self.geodesics_method)

        distances_lists = [list(dl) for dl in distances_lists]  # number_of_boundaries x #V
        self.update_distances_lists(distances_lists)

    def update_distances_lists(self, distances_lists):
        """
        Fills in the distances attributes.
        """
        self._distances_lists = distances_lists
        self._distances_lists_flipped = []  # empty
        for i in range(self.VN):
            current_values = [self._distances_lists[list_index][i] for list_index in range(self.number_of_boundaries)]
            self._distances_lists_flipped.append(current_values)
        self._np_distances_lists_flipped = np.array(self._distances_lists_flipped)
        self._max_dist = np.max(self._np_distances_lists_flipped)

    #  --- Uneven weights
    @property
    def has_uneven_weights(self):
        """ Returns True if the target has uneven_weights calculated, False otherwise. """
        return len(self.weight_max_per_cluster) > 0

    def compute_uneven_boundaries_weight_max(self, other_target):
        """
        If the target has multiple neighborhoods/clusters of vertices, then it computes their maximum distance from
        the other_target. Based on that it calculates their weight_max for the interpolation process
        """
        if self.number_of_boundaries > 1:
            ds_avg_HIGH = self.get_boundaries_rel_dist_from_other_target(other_target)
            max_param = max(ds_avg_HIGH)
            for i, d in enumerate(ds_avg_HIGH):  # offset all distances except the maximum one
                if abs(d - max_param) > 0.01:  # if it isn't the max value
                    ds_avg_HIGH[i] = d + self.offset

            self.weight_max_per_cluster = [d / max_param for d in ds_avg_HIGH]
            logger.info('weight_max_per_cluster : ' + str(self.weight_max_per_cluster))
        else:
            logger.info("Did not compute_norm_of_gradient uneven boundaries, target consists of single component")

    #  --- Relation to other target
    def get_boundaries_rel_dist_from_other_target(self, other_target, avg_type='median'):
        """
        Returns a list, one relative distance value per connected boundary neighborhood.
        That is the average of the distances of the vertices of that boundary neighborhood from the other_target.
        """
        distances = []
        for vi_starts in self.clustered_vkeys:
            ds = [other_target.get_distance(vi) for vi in vi_starts]
            if avg_type == 'mean':
                distances.append(statistics.mean(ds))
            else:  # 'median'
                distances.append(statistics.median(ds))
        return distances

    def get_extreme_distances_from_other_target(self, other_target):
        """
        Returns the minimum and maximum distance of the vertices of this target from the other_target
        """
        extreme_distances = []
        for v_index in other_target.all_target_vkeys:
            extreme_distances.append(self.get_all_distances()[v_index])
        return min(extreme_distances), max(extreme_distances)

    #############################
    #  --- get all distances

    # All distances
    def get_all_distances(self):
        """ Returns the resulting distances per every vertex. """
        return [self.get_distance(i) for i in range(self.VN)]

    def get_all_clusters_distances_dict(self):
        """ Returns dict. keys: index of connected target neighborhood, value: list, distances (one per vertex). """
        return {i: self._distances_lists[i] for i in range(self.number_of_boundaries)}

    def get_max_dist(self):
        """ Returns the maximum distance that the target has on a mesh vertex. """
        return self._max_dist

    #############################
    #  --- per vkey distances

    def get_all_distances_for_vkey(self, i):
        """ Returns distances from each cluster separately for vertex i. Smooth union doesn't play here any role. """
        return [self._distances_lists[list_index][i] for list_index in range(self.number_of_boundaries)]

    def get_distance(self, i):
        """ Return get_distance for vertex with vkey i. """
        if self.has_blend_union:
            return self.blend_union(i)
        else:
            return self.union(i)

    def union(self, i):
        """ Union of distances for vertex with vkey i"""
        return np.min(self._np_distances_lists_flipped[i])

    def blend_union(self, i):
        """ Smooth union of distances for vertex with vkey i"""
        return blend_union_list(values=self._np_distances_lists_flipped[i], r=self.blend_radius)

    #############################
    #  --- scalar field smoothing

    def laplacian_smoothing(self, iterations, strength):
        """ Smooth the distances on the mesh, using iterative laplacian smoothing. """
        L = utils.get_mesh_cotmatrix_igl(self.mesh, fix_boundaries=True)
        new_distances_lists = []

        logger.info('Laplacian smoothing of all distances')
        for i, a in enumerate(self._distances_lists):
            a = np.array(a)  # a: numpy array containing the attribute to be smoothed
            for _ in range(iterations):  # iterative smoothing
                a_prime = a + strength * L * a
                a = a_prime
            new_distances_lists.append(list(a))
        self.update_distances_lists(new_distances_lists)

    #############################
    #  ------ output
    def save_distances(self, name):
        """
        Save distances to json.
        Saves one list with distance values (one per vertex).

        Parameters
        ----------
        name: str, name of json to be saved
        """
        utils.save_to_json(self.get_all_distances(), self.OUTPUT_PATH, name)

    #  ------ assign new Mesh
    def assign_new_mesh(self, mesh):
        """ When the base mesh changes, a new mesh needs to be assigned. """
        mesh.to_json(self.OUTPUT_PATH + "/temp.obj")
        mesh = Mesh.from_json(self.OUTPUT_PATH + "/temp.obj")
        self.mesh = mesh
        self.VN = len(list(self.mesh.vertices()))


####################
#  utils

def blend_union_list(values, r):
    """ Returns a smooth union of all the elements in the list, with blend radius blend_radius. """
    d_result = 9999999  # very big number
    for d in values:
        d_result = blend_union(d_result, d, r)
    return d_result


def blend_union(da, db, r):
    """ Returns a smooth union of the two elements da, db with blend radius blend_radius. """
    e = max(r - abs(da - db), 0)
    return min(da, db) - e * e * 0.25 / r


def chamfer_union(a, b, r):
    """ Returns a chamfer union of the two elements da, db with radius r. """
    return min(min(a, b), (a - r + b) * math.sqrt(0.5))


def stairs_union(a, b, r, n):
    """ Returns a stairs union of the two elements da, db with radius r. """
    s = r / n
    u = b - r
    return min(min(a, b), 0.5 * (u + a + abs((u - a + s) % (2 * s)) - s))


if __name__ == "__main__":
    pass



__all__ = ['layers_default_param']


def layers_default_param(key):
    """ Returns the default parameters with the specified key. """
    if key in default_parameters:
        return default_parameters[key]
    else:
        raise ValueError('The parameter with key : ' + str(key) +
                         ' does not exist in the defaults of curved_slicing parameters. ')


default_parameters = \
    {
        'avg_layer_height': 5.0,
        'max_layer_height': 50.0,
        'min_layer_height': 0.1,
        'vertical_layers_max_centroid_dist': 25.0
    }



"""
********************************************************************************
compas_slicer
********************************************************************************

.. currentmodule:: compas_slicer


.. toctree::
    :maxdepth: 1

    geometry
    slicers
    print_organization
    pre_processing
    post_processing
    utilities

"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division
import os
import compas


__author__ = ['Ioanna Mitropoulou and Joris Burger']
__copyright__ = 'Copyright 2020 ETH Zurich'
__license__ = 'MIT License'
__email__ = 'mitropoulou@arch.ethz.ch'
__version__ = '0.2.1'


HERE = os.path.dirname(__file__)

HOME = os.path.abspath(os.path.join(HERE, "../../"))
DATA = os.path.abspath(os.path.join(HOME, "data"))
DOCS = os.path.abspath(os.path.join(HOME, "docs"))
TEMP = os.path.abspath(os.path.join(HOME, "temp"))

# Check if package is installed from git
# If that's the case, try to append the current head's hash to __version__
try:
    git_head_file = compas._os.absjoin(HOME, '.git', 'HEAD')

    if os.path.exists(git_head_file):
        # git head file contains one line that looks like this:
        # ref: refs/heads/master
        with open(git_head_file, 'r') as git_head:
            _, ref_path = git_head.read().strip().split(' ')
            ref_path = ref_path.split('/')

            git_head_refs_file = compas._os.absjoin(HOME, '.git', *ref_path)

        if os.path.exists(git_head_refs_file):
            with open(git_head_refs_file, 'r') as git_head_ref:
                git_commit = git_head_ref.read().strip()
                __version__ += '-' + git_commit[:8]
except Exception:
    pass

from .geometry import *  # noqa: F401 E402 F403
from .slicers import *  # noqa: F401 E402 F403
from .print_organization import *  # noqa: F401 E402 F403
from .utilities import *  # noqa: F401 E402 F403
from .post_processing import *  # noqa: F401 E402 F403
from .pre_processing import *  # noqa: F401 E402 F403
from .parameters import *  # noqa: F401 E402 F403

__all__ = ["HOME", "DATA", "DOCS", "TEMP"]



import rdp as rdp
import numpy as np
import logging
import progressbar

from compas.geometry import Point

# from compas_slicer.geometry import PrintPoint, Contour

logger = logging.getLogger('logger')

__all__ = ['simplify_paths_rdp']


def simplify_paths_rdp(slicer, threshold):
    """Simplifies a path using the Ramer–Douglas–Peucker algorithm, implemented in the rdp python library.
    https://en.wikipedia.org/wiki/Ramer-Douglas-Peucker_algorithm

    Parameters
    ----------
    slicer: :class:`compas_slicer.slicers.BaseSlicer`
        An instance of one of the compas_slicer.slicers classes.
    threshold: float
        Controls the degree of polyline simplification.
        Low threshold removes few points, high threshold removes many points.

    """

    logger.info("Paths simplification rdp")
    remaining_pts_num = 0

    with progressbar.ProgressBar(max_value=len(slicer.layers)) as bar:
        for i, layer in enumerate(slicer.layers):
            if not layer.is_raft:  # no simplification necessary for raft layer
                for path in layer.paths:
                    pts_rdp = rdp.rdp(np.array(path.points), epsilon=threshold)
                    path.points = [Point(pt[0], pt[1], pt[2]) for pt in pts_rdp]
                    remaining_pts_num += len(path.points)
                    bar.update(i)
        logger.info('%d Points remaining after rdp simplification' % remaining_pts_num)


if __name__ == "__main__":
    pass



import pyclipper
from pyclipper import scale_from_clipper, scale_to_clipper
from compas_slicer.geometry import Layer
from compas_slicer.geometry import Path
from compas.geometry import Point
import compas_slicer
import logging
from compas_slicer.post_processing import seams_align

logger = logging.getLogger('logger')

__all__ = ['generate_brim']


def generate_brim(slicer, layer_width, number_of_brim_offsets):
    """Creates a brim around the bottom contours of the print.

    Parameters
    ----------
    slicer: :class:`compas_slicer.slicers.PlanarSlicer`
        An instance of the compas_slicer.slicers.PlanarSlicer class
    layer_width: float
        A number representing the distance between brim contours
        (typically the width of a layer)
    number_of_brim_offsets: int
        Number of brim paths to add.
    """

    logger.info(
        "Generating brim with layer width: %.2f mm, consisting of %d layers" % (layer_width, number_of_brim_offsets))

    #  TODO: Add post_processing for merging several contours when the brims overlap.
    #  uses the default scaling factor of 2**32
    #  see: https://github.com/fonttools/pyclipper/wiki/Deprecating-SCALING_FACTOR
    SCALING_FACTOR = 2 ** 32

    if slicer.layers[0].is_raft:
        raise NameError("Raft found: cannot apply brim when raft is used, choose one")

    # (1) --- find if slicer has vertical or horizontal layers, and select which paths are to be offset.
    if isinstance(slicer.layers[0], compas_slicer.geometry.VerticalLayer):  # Vertical layers
        # then find all paths that lie on the print platform and make them brim.
        paths_to_offset, layers_i = slicer.find_vertical_layers_with_first_path_on_base()
        for i, first_path in zip(layers_i, paths_to_offset):
            slicer.layers[i].paths.remove(first_path)  # remove first path that will become part of the brim layer
        has_vertical_layers = True

    else:  # Horizontal layers
        # then replace the first layer with a brim layer.
        paths_to_offset = slicer.layers[0].paths
        has_vertical_layers = False

    assert len(paths_to_offset) > 0, 'Attention the brim generator did not find any path on the base. Please check the \
                                      paths of your slicer. '

    # (2) --- create new empty brim_layer
    brim_layer = Layer(paths=[])
    brim_layer.is_brim = True
    brim_layer.number_of_brim_offsets = number_of_brim_offsets

    # (3) --- create offsets and add them to the paths of the brim_layer
    for path in paths_to_offset:
        #  evaluate per path
        xy_coords_for_clipper = []
        for point in path.points:
            # gets the X and Y coordinate since Clipper only does 2D offset operations
            xy_coords = [point[0], point[1]]
            xy_coords_for_clipper.append(xy_coords)

        #  initialise Clipper
        pco = pyclipper.PyclipperOffset()
        pco.AddPath(scale_to_clipper(xy_coords_for_clipper, SCALING_FACTOR), pyclipper.JT_MITER,
                    pyclipper.ET_CLOSEDPOLYGON)

        for i in range(number_of_brim_offsets):
            #  iterate through a list of brim paths
            clipper_points_per_brim_path = []

            #  gets result
            result = scale_from_clipper(pco.Execute((i) * layer_width * SCALING_FACTOR), SCALING_FACTOR)

            for xy in result[0]:
                #  gets the X and Y coordinate from the Clipper result
                x = xy[0]
                y = xy[1]
                z = path.points[0][2]

                clipper_points_per_brim_path.append(Point(x, y, z))

            # adds the first point as the last point to form a closed contour
            clipper_points_per_brim_path = clipper_points_per_brim_path + [clipper_points_per_brim_path[0]]

            #  create a path per brim contour
            new_path = Path(points=clipper_points_per_brim_path, is_closed=True)
            brim_layer.paths.append(new_path)

    brim_layer.paths.reverse()  # go from outside towards the object
    brim_layer.calculate_z_bounds()

    # (4) --- Add the brim layer to the slicer
    if not has_vertical_layers:
        slicer.layers[0] = brim_layer  # replace first layer
    else:
        slicer.layers.insert(0, brim_layer)  # insert brim layer as the first layer of the slicer

    seams_align(slicer, align_with="next_path")


if __name__ == "__main__":
    pass



import logging
import compas
from compas.geometry import Point

logger = logging.getLogger('logger')

__all__ = ['Path']


class Path(object):
    """
    A Path is a connected contour within a Layer. A Path consists of a list of
    compas.geometry.Points.

    Attributes
    ----------
    points: list
        :class:`compas.geometry.Point`
    is_closed: bool
        True if the Path is a closed curve, False if the Path is open.
        If the path is closed, the first and the last point are identical.
    """

    def __init__(self, points, is_closed):
        # check input
        assert isinstance(points[0], compas.geometry.Point)

        self.points = points  # :class: compas.geometry.Point
        self.is_closed = is_closed  # bool

    def __repr__(self):
        no_of_points = len(self.points) if self.points else 0
        return "<Path object with %i points>" % no_of_points

    @classmethod
    def from_data(cls, data):
        """Construct a path from its data representation.

        Parameters
        ----------
        data: dict
            The data dictionary.

        Returns
        -------
        path
            The constructed path.

        """
        points_data = data['points']
        pts = [Point(points_data[key][0], points_data[key][1], points_data[key][2])
               for key in points_data]
        path = cls(points=pts, is_closed=data['is_closed'])
        return path

    def to_data(self):
        """Returns a dictionary of structured data representing the data structure.

        Returns
        -------
        dict
            The path's data.

        """
        data = {'points': {i: point.to_data() for i, point in enumerate(self.points)},
                'is_closed': self.is_closed}
        return data


if __name__ == '__main__':
    pass
